import numpy as np
import pandas as pd

from ..domain import Node


class Recorder():
    r"""A time history recorder

    :param kwargs: additional parameters passed through to specialized recorders.
        See below for more details.

    :Keyword Arguments:
        * **variable** (``list`` of ``str`` or ``All``)
          -- Extra stuff
        * **nodes** (``list`` of ``Node`` ptrs or ``All``)
          -- nodes for which those variables shall be recorded. Use ``nodes=All`` to apply to all nodes.
        * **elements** (``list`` of ``Element`` ptrs or ``All``)
          -- elements for which those variables shall be recorded. Use ``elements=All`` to apply to all elements.
    """

    # hard-coded variable assignments
    system_types = (
        'lam',
        'time',
        'stability',
        'arc-length',
    )
    node_types = (
        'ux',
        'uy',
        'uz',
        'rx',
        'ry',
        'rz',
        'vx',
        'vy',
        'vz',
        'omx',
        'omy',
        'omz',
        'T',
        'stress',
        'strain'
    )
    element_types = (
        'stress',
        'strain',
    )
    material_types = (
        'state',
    )

    def __init__(self, **kwargs):
        self.active   = False
        self.data     = {}

        self.classify(**kwargs)

    def classify(self, **kwargs):

        sysvars      = []
        nodevars     = []
        elemvars     = []
        materialvars = []

        if 'variables' in kwargs:
            for var in kwargs['variables']:
                var_classified = False
                if var in self.system_types or 'type' in kwargs:
                    sysvars.append(var)
                    var_classified = True
                if var in self.node_types:
                    nodevars.append(var)
                    var_classified = True
                if var in self.element_types:
                    elemvars.append(var)
                    var_classified = True
                if var in self.material_types:
                    materialvars.append(var)
                    var_classified = True
                if not var_classified:
                    # this is an unknown/new variable
                    # maybe from a new element type?
                    # just for safety, add it to all types
                    sysvars.append(var)
                    nodevars.append(var)
                    elemvars.append(var)
                    # skip materialvars : too much data

        self.sysvars      = sysvars
        self.nodevars     = nodevars
        self.elemvars     = elemvars
        self.materialvars = materialvars

    def fetchRecord(self, keys=None, source=None):
        r"""
        Request recorded time history data for the listed keys.

        :param keys: If a single key is given as a string, a single `np.array()` is returned.
                     If a list of keys is given, a `list of np.array()` is returned.
        :param source: If a single :py:class:`Node` is given, pick record from that node.
                       If a list of :py:class:`Node` objects is given, match keys and nodes
                       based on order in lists.  **source** and **keys** list must match in shape.
                       If a value should be picked from the model domain instead of a node, enter **None**
                       in the respective slot.
        :returns: time history data for the listed keys.
        """
        if not keys:
            return self.data

        if isinstance(keys, str):
            if source and isinstance(source,Node.Node):
                if source.recorder:
                    return source.recorder.fetchRecord(keys)
                else:
                    return {}
            elif source and (isinstance(source,list) or isinstance(source,tuple)):
                for src in source:
                    if isinstance(src,Node.Node) and src.recorder:
                        return src.recorder.fetchRecord(keys)
                    else:
                        return {}
            else:
                if keys in self.data:
                    return {keys:self.data[keys]}
                else:
                    return {}

        elif isinstance(keys, list) or isinstance(keys, tuple):
            ans = {}

            if source and (isinstance(source,list) or isinstance(source,tuple)):
                for key, src in zip(keys, source):
                    if src and isinstance(src, Node.Node):
                        if src.recorder:
                            node_data = src.recorder.fetchRecord(key)
                            for field in node_data:
                                lbl = node_data[field].label
                                ans[lbl] = node_data[field]
                    else:
                        if key in self.data:
                            ans[key] = self.data[key]
                        else:
                            ans[key] = [0,]

            else:
                for key in keys:
                    if key in self.data:
                        ans[key] = self.data[key]
                    else:
                        ans[key] = [0,]
            return ans
        else:
            return {}

    def addData(self, dta):
        r"""
        *For internal use only.*

        :param dta: variable code as key and scalar value pairs.
        :type dta: dict
        """
        for var in self.data:
            if var in dta:
                self.data[var].data.append(dta[var])
            else:
                self.data[var].data.append(np.nan)
                print(f"Recorder.addData: '{var}' not not in data set: padding with nan")

        for var in dta:
            if var not in self.data:
                print(f"Recorder.addData: '{var}' not initialized by the recorder: ignored")

    def getVariables(self):
        return list(self.data.keys())
    def enable(self):
        """
        Enables the recorder.  You may suspend data collection without loss of previous data by calling ``disable()``

        **Note**: This method should not be called by the user.  Use ``System.startRecorder()`` instead.
        """
        self.active = True

    def disable(self):
        """
        Disables the recorder.  You may resume data collection without loss of previous data by calling ``enable()``

        **Note**: This method should not be called by the user.
        Use ``System.pauseRecorder()`` or  ``System.stopRecorder()`` instead.
        """
        self.active = False

    def isActive(self):
        """
        :returns: **True** if recording is on, False otherwise.
        """
        return self.active

    def reset(self):
        """
        This will reset the data recorder to a *disabled* state and wipe all previously collected data.
        """
        self.active = False
        for var in self.data:
            self.data[var] = []

    def export(self, filename='unknown.txt'):
        """
        :param filename: full path to file where recorded data shall be written to.
            The file type will be determined from the given extension.
            Recognized file types are

            .. list-table::

                * - .txt
                  - tab-separated text file
                * - .csv
                  - comma-separated text file
                * - .hdf
                  - HDF5 file
                * - .json, .jsn
                  - comma-separated text file
                * - .xlsx
                  - MicroSoft Excel file (see the `pandas` package for more detail)
        :type filename: ``str``
        """
        file_parts = filename.split['.']
        if file_parts and len(file_parts)>1:
            suffix = file_parts[-1].lower()
        else:
            suffix = 'txt'   # use text file as a default

        if suffix in ('xlsx',):
            self.export_excel(filename)
        elif suffix in ('csv',):
            self.export_csv(filename)
        elif suffix in ('hdf','hdf5'):
            self.export_hdf(filename)
        elif suffix in ('jsn','json'):
            self.export_json(filename)
        else:
            self.export_text(filename)

    def export_text(self, filename):
        """
        .. warning::

            This method is has not yet been implemented.

        """
        df = pd.DataFrame(self.data)
        df.to_csv(filename, sep='\t', header=True)

    def export_csv(self, filename):
        """
        .. warning::

            This method is has not yet been implemented.

        """
        df = pd.DataFrame(self.data)
        df.to_csv(filename, sep=',', header=True)

    def export_hdf(self, filename):
        """
        .. warning::

            This method is has not yet been implemented.

        """
        df = pd.DataFrame(self.data)
        df.to_hdf(filename)

    def export_json(self, filename):
        """
        .. warning::

            This method is has not yet been implemented.

        """
        df = pd.DataFrame(self.data)
        df.to_json(filename)

    def export_excel(self, filename):
        """
        .. warning::

            This method is has not yet been implemented.

        """
        df = pd.DataFrame(self.data)
        df.to_excel(filename)







"""
Example: plate04

PatchMesher test for the LinearTriangle (Constant Strain Triangle)

"""
from ...examples.Example import *

from ...domain import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.LinearTriangle import *
from ...materials.PlaneStress import *
from ...mesher import *


class ExamplePlate04(Example):

    def docString(self):
        s = """
    ## Patch test for triangular plate under in-plane loading

    The patch test is an empirical minimum test which every finite element has to pass to ensure convergence with mesh refinement.

    It consists of a problem for which a known homogeneous solution exists.  For plates, we commonly use a rectangular plate subject to homogenous edge loading, e.g., constant tension in the x-direction, or constant shear, etc.

    The mesh must contain distorted elements and at least one element not attached to any node on the boundary.
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):
        # ========== setting mesh parameters ==============

        Nx = 6        # number of elements in the mesh
        Ny = 5        # number of elements in the mesh
        Lx = 100.0    # length of plate in the x-direction
        Ly =  80.0    # length of plate in the y-direction

        # ========== setting material parameters ==============

        params = dict(
            E  = 20000.,    # Young's modulus
            nu = 0.250,     # Poisson's ratio
            t  = 1.00       # thickness of the plate
        )

        # ========== setting load parameters ==============

        px  = 10.0         # uniform load normal to x=const
        py  =  0.0         # uniform load normal to y=const
        pxy =  0.0         # uniform shear load on x=const and y=const

        # ========== setting analysis parameters ==============

        target_load_level = 1.00     # reference load
        max_steps = 2                # number of load steps: 2 -> [0.0, 1.0]

        # define a list of target load levels
        load_levels = np.linspace(0, target_load_level, max_steps)

        #
        # ==== Build the system model ====
        #

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        # create nodes

        mesher1 = PatchMesher(model,
                             (0,0), (Lx,0), (Lx,Ly), (0,Ly),                                     # corner nodes
                             (Lx/2, -0.05*Ly), (1.2*Lx, Ly/2), (Lx/2, 0.90*Ly), (0.05*Lx, Ly/2), # mid-side nodes
                             (0.55*Lx, 0.45*Ly))                                                 # center node
        nodes1, elements1 = mesher1.triangleMesh(Nx, Ny, LinearTriangle, PlaneStress(params))

        mesher2 = TriPatchMesher(model,
                                 (0,0), (Lx,0), (Lx/2,Ly),                                       # corner nodes
                                 (Lx/2,0.15*Ly), (0.85*Lx,0.8*Ly), (0.20*Lx,0.6*Ly) # mid-side nodes
                                 )
        mesher2.shift(1.25*Lx, Ly/2)
        #nodes2, elements2 = mesher2.triangleMesh(Nx, LinearTriangle, PlaneStress(params))
        nodes2, elements2 = mesher2.quadMesh(Nx, LinearTriangle, PlaneStress(params))

        nodes    = nodes1    + nodes2
        elements = elements1 + elements2

        # define support(s)

        fix_x = (0,)
        fix_y = (0,4)

        for idx in fix_x:
            nodes[idx].fixDOF('ux')    # horizontal support left end
        for idx in fix_y:
            nodes[idx].fixDOF('uy')          # vertical support right end

        # ==== complete the reference load ====

        # surface loads on the left side
        elements[ 0].setSurfaceLoad(2,px)
        elements[8].setSurfaceLoad(2,px)
        elements[16].setSurfaceLoad(2,px)

        # surface loads on the right side
        elements[ 7].setSurfaceLoad(2,px)
        elements[15].setSurfaceLoad(2,px)
        elements[23].setSurfaceLoad(2,px)

        # these are only nodal forces as part of the reference load
        # .. load only the upper node
        #print(model)

        model.report()

        model.plot(factor=10., title="undeformed system", filename="plate04_undeformed.png")

        model.setLoadFactor(10.0)
        model.solve()

        model.report()

        model.plot(factor=10., filename="plate04_deformed.png")




import numpy as np
from ..LinearElement import *
from ...domain.Node import *
from ...materials.Material import Material

class Triangle(LinearElement):
    r"""
    class: representing a 3-node triangle for diffusion problems
    """

    def __init__(self, node0, node1, node2, material):

        if not material.isMaterialType(Material.DIFFUSION):
            msg = f"Incompatible material type: {self.__class__.__name__} requires type Material.DIFFUSION"
            raise TypeError(msg)

        super().__init__((node0, node1, node2), material)

        self.initialize(
            type=DrawElement.TRIANGLE,
            dofs=['T']
                        )

        # covariant base vectors (reference system)
        base1 = node1.getPos() - node0.getPos()
        base2 = node2.getPos() - node0.getPos()
        self.gcov = np.vstack((base1, base2))

        # metric (reference system)
        self.GIJ = self.gcov @ self.gcov.T

        self.area = np.sqrt(np.linalg.det(self.GIJ)) / 2.0

        # dual base vectors (reference system)
        if np.linalg.det(self.GIJ)<1.0e-10:
            print("Peter, we are having a problem")

        self.gcont = np.linalg.inv(self.GIJ) @ self.gcov

        # dual basis in triangle coordinates
        Gs = self.gcont[0]
        Gt = self.gcont[1]
        Gu = -Gs - Gt
        self.gcont = np.vstack((Gu, self.gcont))

        multiplier  = self.area
        multiplier *= self.material.getThickness()
        multiplier *= self.material.getDiffusivity()

        self.Kt = np.array( [[ Gu @ Gu, Gu @ Gs, Gu @ Gt ],
                             [ Gs @ Gu, Gs @ Gs, Gs @ Gt ],
                             [ Gt @ Gu, Gt @ Gs, Gt @ Gt ]] ) * multiplier


    def __str__(self):
        s = super(Triangle, self).__str__()
        gradPhi = self.material.getGrad()
        s += "\n    grad phi: x={:.3e} y={:.3e}".format(gradPhi[0], gradPhi[1])
        flux = self.material.getFlux()
        s += "\n    flux:     x={:.3e} y={:.3e}".format(flux[0], flux[1])
        if np.array(self.distributed_load).any():
            s += "\n    element flux added to node:"
            for i, P in enumerate(self.Loads):
                Pi = np.array(P) * self.loadfactor
                s += "\n        {}: {}".format(self.nodes[i].getID(), Pi)
        return s

    def setSurfaceLoad(self, face, qn):
        """
        .. list-table::
            :header-rows: 1

            * - face ID
              - nodes defining that face
            * - 0
              - :py:obj:`node 0` to :py:obj:`node 1`
            * - 1
              - :py:obj:`node 1` to :py:obj:`node 2`
            * - 2
              - :py:obj:`node 2` to :py:obj:`node 0`


        :param face: face ID for the loaded face
        :param qn: magnitude of distributed heat flux :math:`q_n`. Influx on a surface is positive.
        """
        if face >= 0 and face <= 2:
            self.faces[face].setLoad(qn, 0.0)

    def resetLoads(self):
        super(Triangle, self).resetLoads()

    def updateState(self):

        # nodal potential/temperature
        T0 = self.getDisp(0,dofs='T')
        T1 = self.getDisp(1,dofs='T')
        T2 = self.getDisp(2,dofs='T')

        # dual base vector in triangle coordinates
        Gu = self.gcont[0]
        Gs = self.gcont[1]
        Gt = self.gcont[2]

        # gradient
        gradT = T0 * Gu + T1 * Gs + T2 * Gt

        # update the material state
        self.material.setGrad(gradT)

        # stress: this is actually the element flux
        flux = self.material.getFlux()
        self.stress = {'Tx':gradT[0], 'Ty':gradT[1], 'qx':flux[0], 'qy':flux[1]}

        multiplier  = self.area
        multiplier *= self.material.getThickness()

        # internal force
        self.Forces = [
            -Gu @ flux * multiplier,
            -Gs @ flux * multiplier,
            -Gt @ flux * multiplier
            ]

        # .. applied element load (reference load)
        self.computeSurfaceLoads()

    def computeSurfaceLoads(self):
        """
        compute surface loads using faces

        This method should be called during :py:meth:`updateState()` by every
        element supporting surface loads

        """
        self.Loads = [ np.zeros_like(self.Forces[I]) for I in range(len(self.nodes)) ]

        for I, face in enumerate(self.faces):
            loads = face.computeNodalFlux()

            # indexing
            J = I+1
            if J>2:
                J -= 3

            # add to element load vectors
            self.Loads[I] += loads[0]
            self.Loads[J] += loads[1]
            if loads.shape[0]>2:
                numNodes = len(self.nodes)
                if numNodes == 6:
                    K = I + 3
                elif numNodes == 8 or numNodes == 9:
                    K = I + 4
                else:
                    msg = "Force data provided from Face2D inconsistent with element data"
                    raise TypeError(msg)

                self.Loads[K] += loads[2]

    def getStress(self):
        return self.Stress

    def mapGaussPoints(self, var):
        r"""
        Initiate mapping of Gauss-point values to nodes.
        This method is an internal method and should not be called by the user.
        Calling that method explicitly will cause faulty nodal values.

        :param var: variable code for a variable to be mapped from Gauss-points to nodes
        """
        # this element has a single gauss-point at s = t = u = 1/3

        flux     = ('qx','qy','Tx','Ty')

        value = 0.0

        if var in flux:
            flow = self.stress
            if var in flow:
                value = flow[var]

        wi     = self.area / 3.
        val_wi = value * wi

        for node in self.nodes:
            node._addToMap(wi, val_wi)









import numpy as np

from ..Element import *
from ...materials.Material import *
from ...domain.Node import *

class Beam2D(Element):
    """
    Representing a 2D beam element.

    **Assumptions**

    * The beam is defined along the X-axis.  (Y-axis values are ignored)
    * Small displacements and small rotations
    * Navier's and Bernoulli-Euler assumptions (plane sections and shear rigidity)
    * Linear elastic material (may change future releases)

    The element is using dofs :math:`v` (:code:`uy`) and :math:`\\theta=v'` (:code:`rz`)

    .. list-table:: Internal variables

        * - self.nodes
          - nodes i and j (tuple)
        * - self.material
          - pointer to Material object
        * - self.force
          - internal force (list of arrays for shear, _V_, and moment, _M_)
        * - self.Forces
          - nodal force vectors (list of np.arrays)
        * - self.Kt
          - tangent stiffness (list of np.arrays)

    """

    def __init__(self, nodei, nodej, material, label=None):
        """
        :param nodei: (pointer to) start Node object
        :param nodej: (pointer to) end Node object
        :param material: a material for axial and flexural integrated behavior
        """
        super().__init__((nodei, nodej), material, label=label)
        self.element_type = DrawElement.CURVE

        if not material.materialType() == Material.SECTION1D:
            raise TypeError("received material type {} but require SECTION1D type".format(material.materialType()))

        dim = nodei.getPos().size

        if dim == 2:
            dof_list = ('uy', 'rz')
        else:
            raise TypeError("spatial dimension of nodes must be 2")

        self._requestDofs(dof_list)

        ndof = len(dof_list)

        # initialize element load to zero
        self.distributed_load = 0.0

        self.L0       = np.linalg.norm(self.nodes[1].getPos() - self.nodes[0].getPos())
        self.force    = 0.0
        self.Forces   = [np.zeros(ndof), np.zeros(ndof)]
        self.Kt       = [[np.zeros((ndof, ndof)), np.zeros((ndof, ndof))],
                         [np.zeros((ndof, ndof)), np.zeros((ndof, ndof))]]
        self.internal_forces = {'fi':0.0, 'Vi':0.0, 'Mi':0.0, 'fj':0.0, 'Vj':0.0, 'Mj':0.0}

    def __str__(self):
        s  = "Beam2D: {} to {}:\n".format( self.nodes[0].getID(), self.nodes[1].getID())
        s += "   material {} properties: {}  strain:{}   stress:{}\n".format(self.material.__class__.__name__,
                                                                           self.material.parameters,
                                                                           self.material.getStrain(),
                                                                           self.material.getStress())
        s += "   nodal forces: Vi:{} Mi:{} Vj:{} Mj:{}".format(*self.Forces[0], *self.Forces[1])
        return s

    def __repr__(self):
        return "Beam2D({},{},{})".format( repr(self.nodes[0]),
                                          repr(self.nodes[1]),
                                          repr(self.material))

    def setDistLoad(self, w):
        """
        :param w: uniform load. positive if pointing in the local y-direction.
        """
        self.distributed_load = w

    def resetLoads(self):
        """

        """
        self.setDistLoad(0.0)
        super(Beam2D, self).resetLoads()

    def getInternalForce(self, variable=''):
        """
        computes vectors of normalized locations (**s**: ndarray) for which
        values (**val**: ndarray) are provided.
        Values of **s** are normalized to the interval :math:`[0,1]`.

        :returns: tuple (s, val)
        """
        self.updateState()

        Xi = self.nodes[0].getPos()
        Xj = self.nodes[1].getPos()
        Nvec = Xj - Xi
        L = np.linalg.norm(Nvec)

        if variable.lower() == 'm' or variable.lower() == 'mz':
            # bending moment (in plane)
            if self.distributed_load:
                s   = np.linspace(0,1,10)
            else:
                s   = np.array([0.,1.])

            Ml  = self.internal_forces['Mi']
            Ml += self.internal_forces['Mw']
            Mr  = self.internal_forces['Mj']
            Mr += self.internal_forces['Mw']

            #val = np.array([Ml, Mr])
            val = Ml * (1. - s) + Mr * s - 0.5 * (self.distributed_load * self.loadfactor) * L*L * s * (1 - s)

        elif variable.lower() == 'v' or variable.lower() == 'vy':
            # transverse shear (in-plane)
            if self.distributed_load:
                s   = np.linspace(0,1,10)
            else:
                s   = np.array([0.,1.])

            Vl  = self.internal_forces['Vi']
            Vl -= self.internal_forces['Pw']
            #Vr  = self.internal_forces['Vj']
            #Vr += self.internal_forces['Pw']

            val = Vl + s * (self.distributed_load * self.loadfactor) * L

        else:
            s   = np.array([0.,1.])
            val = np.zeros_like(s)

        return (s,val)

    def updateState(self):
        """
        Compute internal state, nodal forces, and tangent stiffness for the current state of deformation.
        """

        Xi = self.getPos(0)
        Xj = self.getPos(1)

        # 0 ... uy
        # 1 ... theta
        dispi = self.getDisp(0)
        dispj = self.getDisp(1)

        # compute local coordinate system
        L = Xj[0] - Xi[0]

        # - compute curvature
        """
        Treat flexure are linear elastic for now.
        This is achieved by feeding the material curvature==0
        and, thus, receiving the initial flexural tangent stiffness.
        The moment needs to be computed by alternative means.
        """
        vi     = dispi[0]    # the transformation was performed when we received Ui
        thetai = dispi[1]
        vj     = dispj[0]    # the transformation was performed when we received Uj
        thetaj = dispj[1]

        curvature = 0.0

        # - update material set strain
        self.material.setStrain({'axial':0.0, 'flexure':curvature})

        # - force and moment from the material
        stress = self.material.getStress()
        if 'axial' in stress:
            self.force = stress['axial']
        else:
            self.force = 0.0
        if 'flexure' in stress:
            self.moment = stress['flexure']
        else:
            self.moment = 0.0

        # get tangent moduli from the material
        Et = self.material.getStiffness()
        if 'ax-ax' in Et:
            EA = Et['ax-ax']
        else:
            EA = 0.0
        if 'flx-flx' in Et:
            EI = Et['flx-flx']
        else:
            EI = 0.0
        if ('ax-flx' in Et and Et['ax-flx']) or ('flx-ax' in Et and Et['flx-ax']):
            print('coupling of axial and flexural behavior is ignored by this element')

        #  use the linear stiffness

        kfu = 12. * EI / L**3
        kft =  6. * EI / L**2
        kmu =  kft
        kmt =  4. * EI / L
        kmth = kmt / 2.

        KtII = np.array( [[ kfu,  kft],
                          [ kmu, kmt ]] )
        KtIJ = np.array( [[-kfu,  kft],
                          [-kmu, kmth]] )
        KtJI = np.array( [[-kfu, -kft],
                          [ kmu, kmth]] )
        KtJJ = np.array( [[ kfu, -kft],
                          [-kmu, kmt ]] )

        # nodal forces
        Vi =  kfu * (vi - vj) + kft * (thetaj + thetai)
        Vj = -kfu * (vi - vj) - kft * (thetaj + thetai)

        Mi = kmu * (vi - vj) + kmt * thetai + kmth * thetaj
        Mj = kmu * (vi - vj) + kmth * thetai + kmt * thetaj


        # build element load vector and element tangent stiffness

        # .. internal force
        self.Fi = np.array([Vi, Mi])
        self.Fj = np.array([Vj, Mj])
        self.Forces = (self.Fi, self.Fj)

        # .. tangent stiffness
        self.Kt = np.array( [[KtII, KtIJ],[KtJI, KtJJ]] )

        # internal forces at nodes
        if 'Pw' in self.internal_forces:
            Pw = self.internal_forces['Pw']
        else:
            Pw = 0.0

        if 'Mw' in self.internal_forces:
            Mw = self.internal_forces['Mw']
        else:
            Mw = 0.0
        self.internal_forces = {'fi':self.force, 'Vi': Vi, 'Mi':-Mi,
                                'fj':self.force, 'Vj':-Vj, 'Mj': Mj,
                                'Pw':Pw, 'Mw':Mw}

    def computeSurfaceLoads(self):

        Xi = self.getPos(0)
        Xj = self.getPos(1)

        # compute local coordinate system
        Nvec = Xj - Xi
        L = np.linalg.norm(Nvec)
        Nvec /= L
        Svec = np.array([[0,-1],[1,0]]) @ Nvec

        # .. applied element load (reference load)
        if self.distributed_load:
            Pw = self.distributed_load * L / 2.
            Mw = Pw * L / 6.

            #Pi = np.r_[Pw * Svec, np.array([ Mw])]
            #Pj = np.r_[Pw * Svec, np.array([-Mw])]
            Pi = np.array([Pw, Mw])
            Pj = np.array([Pw, -Mw])
            self.Loads = (Pi, Pj)

            self.internal_forces['Pw'] = Pw * self.loadfactor
            self.internal_forces['Mw'] = Mw * self.loadfactor

    # prepare for removal
    def getAxialForce(self):
        raise DeprecationWarning




import numpy as np
import os
import sys

from .Element import *
from ..recorder.Recorder import Recorder
from .Face2D import *
from .Face3D import *


class LinearElement(Element):
    """
    abstract class: representing a single generic linear element

    The LinearElement adds functions to emulate a nonlinear element
    within the given framework.  Functions include
    * computing the internal force vector for a linear formulation
    *

    """

    def __init__(self, nodes, material, label=None):
        """

        :param nodes:
        :param material:
        :param label:
        """
        super(LinearElement, self).__init__(nodes, material, label=label)

    def getForce(self):
        """
        Request the internal force vector (stress driven force only; **no applied element loads**)

        :return:
        """
        self.updateState()
        return self.Forces





from dis import JUMP_BACKWARD

from fontTools.misc.textTools import caselessSort

from .Mesher import *
from ..domain import Node, System

class PatchMesher(Mesher):
    """
    Mesher for plate and shell elements on a quadrilateral domain,
    including in-plane and out-of plane loaded plates.
    """

    def __init__(self, model, *pts):
        super(PatchMesher, self).__init__(model, *pts)
        if len(pts)<4:
            raise TypeError("insufficient number of points given")

        X = list( pts[:4] )

        if len(pts) > 4 and pts[4]:
            X.append( pts[4] )
        else:
            Xm = 0.5*(np.array(X[0]) + np.array(X[1]))
            X.append( Xm )

        if len(pts) > 5 and pts[5]:
            X.append( pts[5] )
        else:
            Xm = 0.5*(np.array(X[1]) + np.array(X[2]))
            X.append( Xm )

        if len(pts) > 6 and pts[6]:
            X.append( pts[6] )
        else:
            Xm = 0.5*(np.array(X[2]) + np.array(X[3]))
            X.append( Xm )

        if len(pts) > 7 and pts[7]:
            X.append( pts[7] )
        else:
            Xm = 0.5*(np.array(X[3]) + np.array(X[0]))
            X.append( Xm )

        if len(pts) > 8 and pts[8]:
            X.append( pts[8] )
        else:
            Xm = 0.25*(np.array(X[4]) + np.array(X[5]) + np.array(X[6]) + np.array(X[7]))
            X.append( Xm )

        self.X = np.array(X)


    def quadMesh(self, NeX, NeY, element_type, material, **kwargs):
        """
        2D mesher using quadrilateral elements

        :param int NeX: number of elements along the first axis
        :param int NeY: number of elements along the second axis
        :param element_type: compatible element type
        :param material: a material object
        :returns (nodes, elements):
        :var nodes: list of created :py:class:`Node` objects
        :var elements: list of created (subclass of) :py:class:`Element` objects
        """
        # if not element_type.isType(Element.QUAD):
        #     msg="Incompatible element type: {} in {}".format(element_type.__class__.__name__, sys._getframe().f_code.co_name)
        #     raise TypeError(msg)

        match element_type.__name__:
            case 'Quad':
                return self._quad4_mesh(NeX, NeY, element_type, material, **kwargs)
            case 'Quad8':
                return self._quad8_mesh(NeX, NeY, element_type, material, **kwargs)
            case 'Quad9':
                return self._quad9_mesh(NeX, NeY, element_type, material, **kwargs)
            case _:
                msg = "{} does not support element type {}".format(self.__class__.__name__, element_type.__name__)
                raise TypeError(msg)


    def _quad4_mesh(self, NeX, NeY, element_type, material, **kwargs):

        # local coordinates
        ss = np.linspace(-1.,1.,NeX+1)
        tt = np.linspace(-1.,1.,NeY+1)

        nodes    = []
        elements = []

        for t in tt:
            row = []
            for s in ss:
                # mapping from local to global coordinates
                X, Y = self.map(s,t)
                nd = Node(X,Y)
                row.append(nd)
            nodes.append(row)

        for j in range(NeY):
            for i in range(NeX):
                node_list = [ nodes[j][i], nodes[j][i+1], nodes[j+1][i+1], nodes[j+1][i] ]
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)

        nodes = [ nd for row in nodes for nd in row ]

        if self.model and isinstance(self.model, System):
            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        self.nodes    = nodes
        self.elements = elements

        return (nodes, elements)

    def _quad8_mesh(self, NeX, NeY, element_type, material, **kwargs):
        msg = "{} does not yet support element type {}".format(self.__class__.__name__, element_type.__name__)
        raise TypeError(msg)

    def _quad9_mesh(self, NeX, NeY, element_type, material, **kwargs):

        # local coordinates
        ss = np.linspace(-1.,1.,2*NeX+1)
        tt = np.linspace(-1.,1.,2*NeY+1)

        nodes    = []
        elements = []

        for t in tt:
            row = []
            for s in ss:
                # mapping from local to global coordinates
                X, Y = self.map(s,t)
                nd = Node(X,Y)
                row.append(nd)
            nodes.append(row)

        for J in range(0, 2*NeY, 2):
            for I in range(0, 2*NeX, 2):
                node_list = [ nodes[J][I], nodes[J][I+2], nodes[J+2][I+2], nodes[J+2][I],      # corner nodes
                              nodes[J][I+1], nodes[J+1][I+2], nodes[J+2][I+1], nodes[J+1][I],  # midside nodes
                              nodes[J+1][I+1] ]                                                # center node
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)

        nodes = [ nd for row in nodes for nd in row ]

        if self.model and isinstance(self.model, System):
            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        self.nodes    = nodes
        self.elements = elements

        return (nodes, elements)

    def triangleMesh(self, NeX, NeY, element_type, material, **kwargs):
        r"""
        2D mesher using triangular elements

        This is an abstract class - requires implementation in subclass

        :param int NeX: number of elements along the first axis
        :param int NeY: number of elements along the second axis
        :param element_type: compatible element type
        :param material: a material object
        :returns (nodes, elements):
        :var nodes: list of created :py:class:`Node` objects
        :var elements: list of created (subclass of) :py:class:`Element` objects
        """

        match element_type.__name__:
            case 'Triangle':
                return self._triangle3_mesh(NeX, NeY, element_type, material, **kwargs)
            case 'Triangle6':
                return self._triangle6_mesh(NeX, NeY, element_type, material, **kwargs)
            case _:
                msg = "{} does not support element type {}".format(self.__class__.__name__, element_type.__name__)
                raise TypeError(msg)

    def _triangle3_mesh(self, NeX, NeY, element_type, material, **kwargs):

        # local coordinates
        ss = np.linspace(-1.,1.,NeX+1)
        tt = np.linspace(-1.,1.,NeY+1)

        nodes    = []
        elements = []

        for t in tt:
            row = []
            for s in ss:
                # mapping from local to global coordinates
                X, Y = self.map(s,t)
                nd = Node(X,Y)
                row.append(nd)
            nodes.append(row)

        for j in range(NeY):
            for i in range(NeX):
                node_list = [ nodes[j][i], nodes[j][i+1], nodes[j+1][i] ]
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)
                node_list = [ nodes[j+1][i+1], nodes[j+1][i], nodes[j][i+1] ]
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)

        nodes = [ nd for row in nodes for nd in row ]

        if self.model and isinstance(self.model, System):
            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        self.nodes    = nodes
        self.elements = elements

        return (nodes, elements)

    def _triangle6_mesh(self, NeX, NeY, element_type, material, **kwargs):

        # local coordinates
        ss = np.linspace(-1., 1., 2*NeX + 1)
        tt = np.linspace(-1., 1., 2*NeY + 1)

        nodes = []
        elements = []

        for t in tt:
            row = []
            for s in ss:
                # mapping from local to global coordinates
                X, Y = self.map(s, t)
                nd = Node(X, Y)
                row.append(nd)
            nodes.append(row)

        for j in range(0, 2*NeY, 2):
            for i in range(0, 2*NeX, 2):
                node_list = [nodes[j][i], nodes[j][i+2], nodes[j+2][i],           # corner nodes
                             nodes[j][i+1], nodes[j+1][i+1], nodes[j+1][i]]       # midside nodes
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)
                node_list = [nodes[j+2][i+2], nodes[j+2][i], nodes[j][i+2],       # corner nodes
                             nodes[j+2][i+1], nodes[j+1][i+1], nodes[j+1][i+2]]   # midside nodes
                elem = element_type(*node_list, deepcopy(material))
                elements.append(elem)

        nodes = [nd for row in nodes for nd in row]

        if self.model and isinstance(self.model, System):
            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        self.nodes = nodes
        self.elements = elements

        return (nodes, elements)

    def map(self,s,t):
        """
        maps the local coordinates (s,t) from a bi-unit-square
        to the actual position (x,y) in the global model space
        """
        ss = [0.5*s*(s-1), (1-s)*(s+1), 0.5*s*(s+1)]
        tt = [0.5*t*(t-1), (1-t)*(t+1), 0.5*t*(t+1)]
        shp = [ss[0]*tt[0],
               ss[2]*tt[0],
               ss[2]*tt[2],
               ss[0]*tt[2],
               ss[1]*tt[0],
               ss[2]*tt[1],
               ss[1]*tt[2],
               ss[0]*tt[1],
               ss[1]*tt[1] ]
        x = shp @ self.X + self.offset

        return (x[0], x[1])





__all__ = (
    "Integration",
    "LineIntegration",
    "TriangleIntegration",
    "QuadIntegration",
    "TetraIntegration",
    "BrickIntegration",
    "ShapeFunctions",
    "LineShapes",
    "QuadShapes",
    "TriangleShapes",
    "GPdataType"
)

from .LineIntegration import *
from .TriangleIntegration import *
from .QuadIntegration import *
from .TetraIntegration import *
from .BrickIntegration import *

from .LineShapes import *
from .QuadShapes import *
from .TriangleShapes import *

from .GPdataType import *


import numpy as np
import sys


class Faces():
    r"""
    abstract class: representing one face of a 2D or 3D element
    """

    def __init__(self, id, *nds, **kwargs):
        self.id = id
        self.nodes = nds
        self.num_nodes = len(nds)
        self.load = [ 0.0, 0.0 ]   # surface traction in local coordinates: [ pn, ps ] are normal and shear loads
        self.flux = 0.0            # surface flux, out-flux is positive

        if self.num_nodes > 0:
            self.dim = nds[0].getPos().size
        else:
            self.dim = 0

        self.initialize()

    def __str__(self):
        s = "{}_{}".format(self.__class__.__name__, self.id)
        return s

    def initialize(self):
        r"""
        This is a virtual method.  Any class derived from :py:class:`Faces` must implement this function.
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def setLoad(self, pn, ps):
        r"""
        Defines surface loading on this Face.
        Coordinates n and s are outward normal and tangent directions, respectively.
        The local tangent is defined from the first to the second (to the third) node.

        :param pn: normal force per unit length
        :param ps: tangential force per unit length
        """
        self.load = [ pn, ps ]

    def setFlux(self, qn, outflux=False):
        r"""
        Defines a scalar surface flux on this Face.
        Coordinates n is the outward normal.
        **qn** is positive if it is an in-flux

        :param qn: normal flux per unit length (or unit area in 3d)
        :param outflux: set to `True` if given value is an out-flux. (same as entering a negative in-flux)
        """
        if outflux:
            self.flux = -qn  # internally, we always consider qn an in-flux
        else:
            self.flux = qn  # internally, we always consider qn an in-flux

    def computeNodalForces(self):
        r"""
        This is a virtual method.  Any class derived from :py:class:`Faces` must implement this function.
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def computeNodalFlux(self):
        r"""
        This is a virtual method.  Any class derived from :py:class:`Faces` must implement this function.
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)

    def isFace(self, X, N):
        r"""
        Abstract interface for test function.

        .. note::

            This method needs to be implemented by every derived class.

        :param X: position vector for a point
        :type X: np.array
        :param N: outward normal vector at **X**
        :type N: np.array
        :return: **True** if **X** and **N** match this face. **False** otherwise.
        """
        msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
        raise NotImplementedError(msg)



from .DiffusionGeneral import *

class Thermal(DiffusionGeneral):
    """
    Thermal diffusion material description

    .. list-table:: parameters inside the **params** dictionary
       :header-rows: 1

       * - key
         - default value
         - description
       * - **conductivity**
         - 1.0
         - thermal conductivity
       * - **specifict heat**
         - 1.0
         - specific heat capacity
       * - **density**
         - 1.0
         - mass density

    :param params: isotropic thermal properties
    :type params: dict
    """
    
    def __init__(self, params={'conductivity':1., 'specific_heat':1., 'density':1., 'thickness':1.}):

        valid_params = ('conductivity','specific_heat','density','thickness')

        for p in valid_params:
            if p not in params:
                params[p] = 1.0

        for p in params:
            if p not in valid_params:
                msg = f"Warning: unknown parameter {p} found in parameter list - ignored"
                print(msg)

        general_params = dict(
            diffusivity = params['conductivity'],
            capacity    = params['specific_heat'],
            density     = params['density'],
            thickness   = params['thickness']
        )
        super(Thermal, self).__init__(params=general_params)



from multiprocessing.connection import answer_challenge
from xmlrpc.client import INVALID_ENCODING_CHAR

import numpy as np

###
### matplotlib is needed for deprecated methods only
###
### remove this section onve ALL plotting has been moved into suitable plotter classes
###
import matplotlib.pyplot as plt
from fontTools.misc.textTools import caselessSort

###
### end of deprecated section
###

from .Node import Node
from ..elements.Element import *
from ..solver.LinearSolver import LinearSolver
from ..plotter.ElementPlotter import ElementPlotter as Plotter
from ..recorder import *


class System():
    """
    class: representing a System model
    """

    def __init__(self, verbose=False):
        self.nodes       = []
        self.elements    = []
        self.constraints = []
        self.plotter     = Plotter()

        self.verbose = verbose

        self.disp        = np.array([])
        self.loads       = np.zeros_like(self.disp)

        # global analysis settings
        self.loadfactor = 1.0

        self.solver = LinearSolver()
        self.solver.connect(self, self.nodes, self.elements, self.constraints)

        self.initRecorder()
        self.trackStability(False)

    def __str__(self):
        s = "System object"
        for node in self.nodes:
            s += "\n" + repr(node)
        for elem in self.elements:
            s += "\n" + repr(elem)
        return s

    def __repr__(self):
        return "System()"

    def setSolver(self, solver):
        """
        This method will change the current solver to the provided solver.

        Upon successful update of the solver, the old solver state will be
        pushed to the new solver.

        :param solver: a pointer to a solver object.
        """
        state = {}
        if self.solver:
            state = self.solver.fetchState()
        if solver:
            self.solver = solver
            self.solver.pushState(state)

    def addNode(self, *nodes):
        """
        Add one or more nodes to the model.

        :param newNode: a :code:`Node` object
        """
        for newNode in nodes:
            if newNode not in self.nodes:
                newNode.setLoadFactor(self.loadfactor)
                self.nodes.append(newNode)
            elif self.verbose:
                print('addNode: node {} already exists in system and was not added again'.format(newNode.getID()))

    def __add__(self, other):
        if isinstance(other, Node):
            self.addNode(other)
        elif isinstance(other, Element):
            self.addElement(other)
        else:
            raise TypeError
        return self

    def addElement(self, *newElements):
        """

        :param newElement: an :code:`Element` object
        """
        for elem in newElements:
            elem.setLoadFactor(self.loadfactor)
            self.elements.append(elem)

    def addConstraint(self, *newConstraints):
        """

        :param newConstraints: one or more :py:class:`Constraint` or a subclass objects
        """
        for constraint in newConstraints:
            self.constraints.append(constraint)

# --------- load control functions ----------------------

    def setLoadFactor(self, lam):
        self.solver.setLoadFactor(lam)

    def resetDisplacements(self):
        self.solver.resetDisplacements()

    def setConstraint(self, nodeIdx, nvec, Ubar):

        self.en = np.zeros(self.sdof)
        idx = np.arange(nodeIdx*self.ndof, (nodeIdx+1)*self.ndof)
        self.en[idx] = nvec
        self.targetU = Ubar

        self.hasConstraint = True
        self.useArcLength = False

    def getSolver(self):
        r"""
        Provides a pointer to the current solver instance.

        This function is used to get access to the solver and give instructions directly to that solver.
        """
        return self.solver

# --------- modeling assist functions ----------------

    def _is_node_on_line(self, node, pos, dir, tol=1.e-3):
        tvec = np.array(dir)
        tvec /= np.linalg.norm(tvec)  # normalized tangent vector

        R = node.getPos() - np.array(pos)
        distX = tvec @ R
        distR = R - distX * tvec
        dist = np.linalg.norm(distR)

        return (dist < tol)


    def findNodesAt(self, pos, tol=1.e-3):
        r"""
        Find any node within distance tol from the given target position (**pos**).

        If a 2d position is given for a 3d-problem, the input target defines a cylinder with a vertical axis (3rd direction) and radius **tol**.

        If a 3d position is given for a 2d-problem, the 3rd coordinate of **pos** will be ignored.

        :param pos: a coordinate tuple (2d or 3d) defining th etarget position
        :param tol: search radius for find. Any node within distance **tol** will be returned.
        :return: a list of tuples (node_pointer, distance)
        """
        ans = []

        for node in self.nodes:
            dist = node.distanceTo(pos)
            if dist<tol:
                ans.append((node,dist))

        return ans

    def findNodesAlongLine(self, pos, dir, tol=1.e-3):
        r"""
        Find any node within distance tol from the given target position (**pos**).

        If a 2d position is given for a 3d-problem, the input target defines a cylinder with a vertical axis (3rd direction) and radius **tol**.

        If a 3d position is given for a 2d-problem, the 3rd coordinate of **pos** will be ignored.

        :param pos: a coordinate tuple (2d or 3d) defining th etarget position
        :param dir: direction vector defining the orientation of the line
        :param tol: search radius for find. Any node within distance **tol** will be returned.
        :return: a list of tuples (node_pointer, distance)
        """
        ans = []

        for node in self.nodes:
            if self._is_node_on_line(node, pos, dir, tol=tol):
                ans.append((node, node.distanceTo(pos)))

        return ans

    def findNodesOnPlane(self, pos, dir, tol=1.e-3):
        r"""
        Find any node within distance tol from the reference plane.

        The plane is defined by

        .. math::
            | {\bf n} \cdot ( {\bf X}_{node} - {\bf X}_0 ) | < TOL

        where
        :math:`{\bf n}` is the unit normal vector to the plane (normalized **dir**),
        :math:`{\bf X}_{node}` is the position vector of any node, and
        :math:`{\bf X}_0 )` is the position vector of a reference point on the plane (**pos**),

        :param pos: a coordinate tuple (2d or 3d) defining th etarget position
        :param dir: normal vector to the target plane
        :param tol: search radius for find. Any node within distance **tol** will be returned.
        :return: a list of tuples (node_pointer, distance)
        """
        X0 = np.array(pos)             # reference point
        nvec = np.array(dir)
        nvec /= np.linalg.norm(nvec)   # normalized normal vector

        ans = []

        for node in self.nodes:
            dist = np.abs( nvec @ (node.getPos() - X0) )
            if dist<tol:
                ans.append((node,dist))

        return ans

    def findFacesAlongLine(self, pos, dir, orientation=0, tol=1.e-2):
        r"""
        Find any element faces within distance tol from the given target position (**pos**).
        If a 2d position is given for a 3d-problem, the input target defines a cylinder with a vertical axis (3rd direction) and radius **tol**.
        If a 3d position is given for a 2d-problem, the 3rd coordinate of **pos** will be ignored.


        .. list-table::
            :header-rows: 1

            * - orientation
              - filter
            * - `0`
              - Any face aligned with the plane will be returned
            * - `+1`
              - Only faces with an outward normal pointing to the right of **dir** will be selected.
            * - `-1`
              - Only faces with an outward normal pointing to the left of **dir** will be selected.



        :param pos: a coordinate tuple (2d or 3d) defining th etarget position
        :param dir: direction vector defining the orientation of the line
        :param orientation: search filter
        :param tol: search radius for find. Any node within distance **tol** will be returned.
        :return: a list of tuples (node_pointer, distance)
        """
        ans = []

        Xo = np.array(pos)
        To = np.array(dir)
        To /= np.linalg.norm(To)

        match len(dir):
            case 2:
                No = np.array([[ 0.0, 1.0],[ -1.0, 0.0]]) @ To  # rotate to the right
            case 3:
                msg = "findFacesAlongLine requires a 2d-analysis.  Use findEdgesAlongLine() for shells and findFacesOnPlane() for 3d solids."
                raise TypeError(msg)

        boundary_nodes = self.findNodesAlongLine(pos, dir, tol=tol)

        for node, dnode in boundary_nodes:
            for elem in node.elements:
                #print('+', elem)
                for face in elem.faces:
                    for x, area in zip(face.pos, face.area):
                        match orientation:
                            case 0:   # both sides
                                dist      = np.abs((x - Xo) @ No)
                                area_test = np.abs( No @ area / np.linalg.norm(area) )
                            case -1:  # left side only
                                dist      = -(x - Xo) @ No
                                area_test = -No @ area / np.linalg.norm(area)
                            case 1:   # right side only
                                dist      = (x - Xo) @ No
                                area_test = No @ area / np.linalg.norm(area)

                        if  dist < tol and area_test > (1. - 10. * tol):
                            # avoid duplication
                            if (elem,face) not in ans:
                                ans.append((elem,face))

        return ans

    def findFacesOnPlane(self, pos, dir, orientation=0, tol=1.e-3):
        r"""
        Find any element faces within distance tol from the given target position (**pos**).

        If a 2d position is given for a 3d-problem, the input target defines a cylinder with a vertical axis (3rd direction) and radius **tol**.

        If a 3d position is given for a 2d-problem, the 3rd coordinate of **pos** will be ignored.


        .. list-table::
            :header-rows: 1

            * - orientation
              - filter
            * - `0`
              - Any face aligned with the plane will be returned
            * - `+1`
              - Only faces with an outward normal pointing in the direction of **dir** will be selected.
            * - `-1`
              - Only faces with an outward normal pointing against the direction of **dir** will be selected.


        :param pos: a coordinate tuple (2d or 3d) defining th etarget position
        :param dir: normal vector to the target plane
        :param orientation: search filter
        :param tol: search radius for find. Any node within distance **tol** will be returned.
        :return: a list of tuples (element_ptr, face_ptr)
        """
        ans = []
        return ans

# --------- displacement control functions ----------------
    def setDisplacementControl(self, node, dof, target):
        r"""
        activate displacement control for the next load step

        :param node:     pointer to the controlling node
        :param dof:      dof code for controlled dof
        :param target:   target displacement value
        """
        if self.solver:
            self.solver.setDisplacementControl(node, dof, target)

# --------- Arc-length control functions: forward to Solver ------------

    def initArcLength(self, load_increment=1., alpha=0.0, tolerance=1.0e-12):
        r"""
        Initializes parameters for the arc-length constraint.

        .. math::

           g({\bf u}, \lambda)) := \alpha ||\bar {\bf P}|| (\lambda-\lambda_n)^2 + ({\bf u} - {\bf u}_n)({\bf u} - {\bf u}_n) - \Delta s^2 = 0

        .. note::

           This feature requires a nonlinear solver. Review the :py:meth:`setSolver` function.

        :param load_increment:   load increment used to calibrate the constraint
        :param alpha:            load contribution factor
        :param tolerance:        convergence tolerance
        """
        if self.solver:
            self.solver.initArcLength(load_increment=load_increment, alpha=alpha, tolerance=tolerance)

    def stepArcLength(self, verbose=False, max_iter=10):
        r"""
        Progresses the model state by one arc-length.

        .. note::

            You need to initialize arc-length control by one call to
            :py:meth:`initArcLength` at least once to set all necessary parameters.

        :return normR: the norm of the generalized residuum from the last iteration step
        """

        if self.solver:
            loadfactor, normR = self.solver.stepArcLength(verbose=verbose, max_iter=max_iter)
            self.setLoadFactor(loadfactor)

    # --------- recorder methods ------------------------------

    def initRecorder(self, **kwargs):
        r"""
        initializes data arrays for gathering of load history data

        :keyword variables: list of variables or d.o.f.-codes to be recorded
        :type variables: list-type
        :keyword nodes:  nodes for which to record
        :type nodes: list of :py:class:`Node`
        :keyword elements:  elements for which to record
        :type elements: list of :py:class:`Element`

        .. code::

            # examples:

            # system recorder tracking the stability index
            initRecorder(variables=['stability',])

            # tracking 'ux' at nodes X1 and X7
            initRecorder(variables=['ux',], nodes=[X1, X7])

            # tracking all components of stress at Elem2 and Elem42
            initRecorder(variables=['stress',], elements=[Elem2, Elem42, ...])

        """
        if 'variables' in kwargs:
            if 'lam' not in kwargs['variables']:
                kwargs['variables'] = list(kwargs['variables']) + ['lam']
            #if 'stability' not in kwargs['variables']:
            #    kwargs['variables'] = list(kwargs['variables']) + ['stability']
        else:
            kwargs['variables'] = ['lam','stability']

        if 'nodes' in kwargs and isinstance(kwargs['nodes'],str):
            if kwargs['nodes'].lower() == 'all':
                kwargs['nodes'] = self.nodes
            else:
                msg = f"Unknown node identifier `nodes={kwargs['nodes']} in Recorder`"
                raise KeyError(msg)

        if 'elements' in kwargs and isinstance(kwargs['elements'],str):
            if kwargs['elements'].lower == 'all':
                kwargs['elements'] = self.elements
            else:
                msg = f"Unknown node identifier `elements={kwargs['elements']} in Recorder`"
                raise KeyError(msg)

        self.recorder = ModelRecorder(**kwargs)

    def startRecorder(self):
        """
        starts the recorder
        """
        if self.recorder:
            self.recorder.enable()
        if self.solver:
            self.solver.startRecorder()
        for node in self.nodes:
            node.startRecorder()
        for elem in self.elements:
            elem.startRecorder()

    def pauseRecorder(self):
        """
        pauses the recorder.  You can restart the recorder by calling
        `startRecorder()`.
        """
        if self.recorder:
            self.recorder.disable()
        if self.solver:
            self.solver.pauseRecorder()
        for node in self.nodes:
            node.pauseRecorder()
        for elem in self.elements:
            elem.pauseRecorder()

    def stopRecorder(self):
        """
        stops the recorder
        """
        if self.recorder:
            self.recorder.disable()
        if self.solver:
            self.solver.stopRecorder()
        for node in self.nodes:
            node.stopRecorder()
        for elem in self.elements:
            elem.startRecorderstopRecorder()

    def recordThisStep(self):
        """
        record current state of the system
        """

        for node in self.nodes:
            node.recordThisStep(self.loadfactor)

        for elem in self.elements:
            elem.recordThisStep(self.loadfactor)

        if self.recorder and self.recorder.isActive():
            data = {'lam':self.loadfactor}
            if self.track_stability:
                data['stability'] = self.solver.checkStability(num_eigen=1)
            else:
                data['stability'] = np.nan

            self.recorder.addData(data)

    def trackStability(self, on=True):
        self.track_stability = on

    def fetchRecord(self, keys=[]):
        """
        Request recorded time history data for the listed keys.
        If a single key is given as a string, a single `np.array()` is returned.
        If a list of keys is given, a `list of np.array()` is returned.

        :returns: time history data for the listed keys.
        """
        if keys:
            return self.recorder.fetchRecord(keys=keys)
        else:
            return []

# ------- solver methods ----------

    def solve(self, **kwargs):
        """
        Solve system of equations and find state of deformation for the given load level.
        """
        if self.solver:
            self.solver.solve(**kwargs)
            if self.solver.hasConstraint:
                # spread the news about the new load level throughout the system
                self.setLoadFactor(self.solver.loadfactor)
        else:
            msg = "** WARNING ** {}.{} not implemented".format(self.__class__.__name__, sys._getframe().f_code.co_name)
            raise NotImplementedError(msg)

    def checkStability(self, **kwdargs):
        r"""
        Computes the stability index as

        * :math:`\mathop{det}([{\bf K}_t])` for systems with less than 25 d.o.f.s
        * :math:`\min\lambda_i` where :math:`\lambda_i` are the eigenvalues of :math:`{\bf K}_t`

        **Implemented** by :py:class:`Solver`.

        :param verbose: set to **True** for log info
        :param num_eigen: if set to a value greater than 0, show the **num_eigen** eigenvalues
                        closest to the current load level.
        :returns: stability index
        """
        if self.solver:
            return self.solver.checkStability(**kwdargs)

    def getBucklingMode(self, mode=0, **kwargs):
        if self.solver:
            return self.solver.getBucklingMode(mode=mode, **kwargs)
        else:
            return (np.nan,None)

# ------------ plot methods -----------------

    def plot(self, factor=1.0, show_reactions=True, show_loads=True, force_limit=1.0e-6, **kwargs):
        """
        Create mesh plot showing the undeformed and the deformed system

        If **filename** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param factor: deformation magnification factor
        :param show_reactions: plot reaction forces if this evaluates to **True**
        :param show_loads:  plot nodal load vectors if this evaluates to **True**
        :param force_limit: don't plot forces smaller than this value
        :param filename:  filename (str)
        """

        self.plotter.setMesh(self.nodes, self.elements)

        if show_loads:

            kwargs['show_loads'] = True
            R = self.solver.getNodalLoads(dofs=('ux','uy','rz'), cut_off=force_limit)
            self.plotter.setNodalLoads(R)

        if show_reactions:

            kwargs['show_reactions'] = True
            R = self.solver.getNodalReactions(dofs=('ux','uy','rz'), cut_off=force_limit)
            self.plotter.setReactions(R)

        self.plotter.displacementPlot(factor=factor, **kwargs)

    def valuePlot(self, variable, factor=0.0, filename=None, **kwargs):
        r"""
        Create a false color contour plot for the selected variable.
        A value of zero (0.0) will be assigned for any variable not
        provided by an element or a node.

        .. list-table:: known variables
            :header-rows: 1

            * - keyword
              - description
            * - 'ux', 'uy', 'uz'
              - displacement component in the `x`, `y` or `z` direction
            * - 'rx', 'ry', 'rz'
              - rotation around the `x`, `y` or `z` axis
            * - 'sxx', 'syy', 'szz'
              - normal stress :math:`\sigma_{xx}`, :math:`\sigma_{yy}`, :math:`\sigma_{zz}`
            * - 'sxy', 'syz', 'szx'
              - shear stress :math:`\sigma_{xy}`, :math:`\sigma_{yz}`, :math:`\sigma_{zx}`
            * - 'epsxx', 'epsyy', 'epszz'
              - normal strain :math:`\varepsilon_{xx}`, :math:`\varepsilon_{yy}`, :math:`\varepsilon_{zz}`
            * - 'epsxy', 'epsyz', 'epszx'
              - engineering shear strain :math:`\gamma_{xy}=2\varepsilon_{xy}`,
                :math:`\gamma_{yz}=2\varepsilon_{yz}`,
                :math:`\gamma_{zx}=2\varepsilon_{zx}`
            * - 'nxx', 'nyy', 'nxy'
              - membrane forces :math:`n_{xx}=h\sigma_{xx}`, :math:`n_{yy}=h\sigma_{yy}`, :math:`n_{xy}=h\sigma_{xy}`
            * - 'T'
              - temperature
            * - 'qx', 'qy', 'qz'
              - `x`, `y` or `z` component of the temperature gradient, :math:`q_i = \partial T/\partial x_i`


        :param variable: string code for variable to show
        :param factor: deformation magnification factor (default is undeformed)
        :param filename: if **filename** is given, store the plot to that file.
                Use proper file extensions to indicate the desired format (.png, .pdf)
        :type filename: str
        """
        varIsAtGaussPoint = True

        for node in self.nodes:
            if node.hasDOF(variable):
                varIsAtGaussPoint = False
                break

        if varIsAtGaussPoint:
            for node in self.nodes:
                node._resetGaussPointMap(variable)
            for elem in self.elements:
                elem.mapGaussPoints(variable)

        self.plotter.setMesh(self.nodes, self.elements)
        self.plotter.valuePlot(variable_name=variable,
                               factor=factor,
                               gaussptvalue=varIsAtGaussPoint,
                               filename=filename, **kwargs)

    def beamValuePlot(self, variable, factor=0.0, filename=None, **kwargs):
        """
        Create a traditional beam value plot, i.e., moment and shear diagrams.


        :param variable: string code for variable
        :param deformed: True | **False**
        :param filename: if **filename** is given, store the plot to that file.
                Use proper file extensions to indicate the desired format (.png, .pdf)
        :type filename: str
        """

        self.plotter.setMesh(self.nodes, self.elements)
        self.plotter.beamValuePlot(variable_name=variable, factor=factor, filename=filename, **kwargs)

    def historyPlot(self, varX='', varY='', filename=None, **kwargs):
        """
        Create a generic X-Y plot using recorder data for load-level (horizontal)
        and varY (vertical).

        If **filename** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param varX:     X-axis parameter;
                         This must be a variable code previously set by :py:meth:`initRecorder`.

                         If :py:attr:`varX` is a :py:class:`str`, it is assumed to be a global parameter/variable, e.g.,
                         load factor **lam**, or pseudo time **time** (time not yet supported)

                         If :py:attr:`varX` is a :py:class:`tuple` containing a variable code and a :py:class:`Node` pointer,
                         the respective nodal DOF defined by the parameter will be used.

                         If :py:attr:`varX` is a :py:class:`tuple` containing a variable code and a :py:class:`Element` pointer,
                         the respective element DOF defined by the parameter will be used. The parameter must identify a scalar parameter.
                         Hence, you may use :py:data:`'sig:xx'` but **not** :py:data:`'stress'`.

        :param varY:     a variable code previously set by :py:meth:`initRecorder`
        :param filename:
        :type filename: str
        :param node: (optional)
        :param nodes: (optional)
        """
        if not self.recorder:
            return

        if not varX:
            variableX = 'lam'
            src = [None]
        elif isinstance(varX, list) or isinstance(varX, tuple):
            variableX = varX[0]
            src = [varX[1]]
        else:
            variableX = varX
            src = [None]

        if isinstance(varY, list) or isinstance(varY, tuple):
            variableY = varY
        else:
            variableY = [varY]

        if 'nodes' in kwargs:
            data = self.recorder.fetchRecord([variableX, *variableY], source=src + kwargs['nodes'])
        elif 'node' in kwargs:
            data = self.recorder.fetchRecord([variableX, *variableY], source=src + [kwargs['node']])
        elif src:
            data = self.recorder.fetchRecord([variableX, *variableY], source=src + [None]*len(variableY))
        else:
            data = self.recorder.fetchRecord([variableX, *variableY])


        X = Record()
        for key, record in data.items():
            if record.key == variableX:
                X = record

        Y = []
        for key, record in data.items():
            if record.key in variableY:
                Y.append( record )

        if 'title' not in kwargs:
            kwargs['title'] = "Load History Plot"

        self.plotter.xyPlot(X, Y, filename=filename, **kwargs)

    def plotDOF(self, dofs=[]):
        """
        .. warning::

            This method has been marked **DEPRECATED** !

            Use the :code:`System.plot()` method with appropriate :code:`**kwargs` instead.
        """


        if self.track_stability:
            fig, (ax1, ax2) = plt.subplots(1,2, figsize=(10,4))
        else:
            fig, ax1 = plt.subplots(1,1, figsize=(5,4))

        if not dofs:
            dofs = range(self.sdof)

        loadfactors   = self.recorder['lambda']
        displacements = np.array(self.recorder['U'])

        for idx in dofs:

            nodeID    = idx // self.ndof + 1
            direction = idx % self.ndof

            if direction==0:
                style = '--x'
                lbl = '$ u_{{{}}} $'.format(nodeID)
            elif direction==1:
                style = '-o'
                lbl = '$ v_{{{}}} $'.format(nodeID)
            elif direction==2 and direction < self.ndim:
                style = ':+'
                lbl = '$ w_{{{}}} $'.format(nodeID)
            else:
                if direction < self.ndim:
                    style = '-+'
                    lbl = '$ u_{{{},{}}} $'.format(nodeID,direction)
                else:
                    style = '-.o'
                    if self.ndim < 3:
                        lbl = '$ \\theta_{{{}}} $'.format(nodeID)
                    else:
                        lbl = '$ \\theta_{{{},{}}} $'.format(nodeID,direction-self.ndim)

            ax1.plot(displacements[:,idx],loadfactors, style,label=lbl)

        ax1.grid(True)
        ax1.set_xlabel('component of displacement, $ u_i$')
        ax1.set_ylabel('load factor, $\\lambda$')
        ax1.legend()

        if self.track_stability:

            detKt = np.array(self.recorder['stability'])

            ax2.plot(detKt, loadfactors,'-r')
            ax2.plot(np.zeros_like(loadfactors), loadfactors,'-k', lw=2)
            ax2.grid(True)
            ax2.set_xlabel('stability index, $ det({\\bf K}_t) $ or $ \\min |\\lambda| $')
            ax2.set_ylabel('load factor, $\\lambda$')

        plt.savefig("history_plots.png", bbox_inches='tight')
        plt.show()

    def plotSystem(self, filename=""):
        r"""
        This is a shortcut for

        .. code::

            System.plot(filename=filename,
                        factor=0.0,
                        title="Undeformed system",
                        show_loads=1,
                        show_reactions=0)


        :param filename: if given, save the plot to that file
        """
        self.plot(filename=filename, factor=0.0, title="Undeformed system", show_loads=1, show_reactions=0)

    def pushU(self):
        """
        Store the current displacement vector for later restore using :code:`popU()`.
        """
        for node in self.nodes:
            node.pushU()

    def popU(self):
        """
        Restore a previously pushed displacement vector (using :code:`pushU()`).
        """
        for node in self.nodes:
            node.popU()

    def plotBucklingMode(self, factor=1.0, mode=0, filename=None, **kwargs):
        """
        .. warning::

            This method has been marked **DEPRECATED** !

            Use the :code:`System.plot()` method with appropriate :code:`**kwargs` instead.

        Select a mode shape by setting `mode` to an integer, where `0` stands for the mode with the smallest
        absolute eigenvalue, `1` the second smallest, and so forth.  If the resulting eigenvalue (shown in the plot)
        is negative, the load-level associated with that mode is below the current load level.

        If a `filename` is given, e.g., `filename="buckling_mode.png"`, the plot will be saved to the respective file.
        The file type will be determined from the given extensions, e.g., `.png` for a PNG format.

        :param factor: Scale the mode shape by this factor.
        :param mode:   which mode shape shall be plotted?
        :param filename: string
        :param kwargs: not specified here.  These are handed over to the :code:`Solver(...)` instance.
        """

        # perform eigenvalue analysis
        lam = self.getBucklingMode(mode=mode, **kwargs)
        title_text = f"Mode Shape for $ \\lambda = {lam:.2f} $"
        self.plot(factor=factor, filename=filename, title=title_text, modeshape=True)

    def report(self):
        """
        print a text-based summary report

        """
        s  = "\nSystem Analysis Report\n"
        s += "=======================\n"
        s += "\nNodes:\n"
        s += "---------------------\n"
        for node in self.nodes:
            for ln in str(node).split('\n'):
                s += "  " + ln + "\n"
        s += "\nElements:\n"
        s += "---------------------\n"
        for elem in self.elements:
            for ln in str(elem).split('\n'):
                s += "  " + ln + "\n"
        print(s)

# ------------ operational support methods --------------

    def setLoadFactor(self, lam):
        r"""
        Set the target load factor to **lam**

        The entered load pattern is considered a reference load,
        to be multiplied by a scalar load factor, :math:`\lambda`.

        If no load factor is set explicitly, a factor of 1.0 is assumed, i.e., the
        entire entered load is applied in full.
        """
        self.loadfactor = lam

        # inform elements about the load factor
        for elem in self.elements:
            elem.setLoadFactor(lam)

        # inform nodes about the new load factor
        for node in self.nodes:
            node.setLoadFactor(lam)

        # let solver know about the new load factor
        if self.solver:
            self.solver.setLoadFactor(lam)

    def resetDisp(self):
        """
        Resets the displacement vector.
        """
        for node in self.nodes:
            node.resetDisp()

    def resetLoad(self):
        """
        Resets the load vector.
        """
        for node in self.nodes:
            node.resetLoad()

    def resetAll(self):
        """
        Resets load and displacement vectors.
        """
        self.resetDisp()
        self.resetLoad()



if __name__ == "__main__":

    msg = """
    This file is not an executable script.
    See the FEM.edu manual for details on how to run the program.
    """
    print(msg)




class Constraint():
    """
    The head parent for all constrint implementations
    """

    COUNT = 0

    def __init__(self, nodes_list, params, **kwargs):
        self.id = self.COUNT  # creating a unique ID
        self.COUNT += 1       # ensuring the next ID will be unique as well
        self.start = -1       # index in system matrix

        self.numnds = len(nodes_list)
        self.nodes  = nodes_list
        self.params = params

        self.neqns  = 0

    def __str__(self):
        s = "{}: ".format(self.__class__.__name__)
        for node in self.nodes:
            s += "{} ".format(node.getID())
        return s

    def __repr__(self):
        return str(self)

    def countConditions(self):
        """
        :returns: the count of constraint equations in this Constraint object
        """
        return self.neqns

    def on_converged(self):
        pass

    def revert(self):
        pass




"""
documentation string inside `__init__.py`
"""

__all__ = (
    "Triangle",
    "Triangle6",
    "Quad",
    "ReducedIntegrationQuad",
    "HRQuad",
    "Quad8",
    "Quad9",
    "Truss",
    "Spring",
    "Beam2D",
    "Frame2D",
    "Faces",
    "Face2D",
    "Face3D",
)

from .Beam2D import *
from .Quad import *
from .Quad9 import *
from .ReducedIntegrationQuad import *
from .HRQuad import *
from .Triangle import *
from .Triangle6 import *
from .Truss import *
from .Spring import *
from .Frame2D import *



"""
Example: frame02

Buckling of a vertical beam with pin-pin support

modeled using a 2D frame element

.. code::

    |
    | P
    v
    x o ... support
    I
    x   ... node
    I
    x   ... node
    ^   ... support

    x ..... node
    I ... frame element
    <-- ... applied force
    ^ ..... pin support
    o ..... roller support

    degrees of freedom:
    0 ... horizontal displacement
    1 ... vertical displacement
    2 ... rotation, theta


.. list-table:: setting given parameters

    * - N  = 2
      - number of elements
    * - L  = 100.0
      - column length
    * - EA = 2000000.0
      - axial stiffness
    * - EI = 21000.0
      - flexural stiffness
    * - w  = 0.1
      - applied lateral load


"""
from ...examples.Example import *

from ...domain import *
from ...solver.NewtonRaphsonSolver import *
from ...elements.Frame2D import *
from ...materials.ElasticSection import *


class ExampleFrame02(Example):

    def docString(self):
        s = """
    Buckling of a beam with pin-pin support
    
    modeled using a 2D frame element
        
        |
        | P
        v
        x o ... support
        I
        x   ... node
        I
        x   ... node
        ^   ... support
    
        x ..... node
        I ... frame element
        <-- ... applied force
        ^ ..... pin support
        o ..... roller support

    degrees of freedom:
        ux ... horizontal displacement
        uy ... vertical displacement
        rz ... rotation, theta
        
    parameters 
        
        N  = 2     # number of elements
        L  = 100.0
        EA = 2000000.0
        EI = 21000.0
        w  = 0.1
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):
        # initialize a system model

        N  = 8     # number of elements
        L  = 100.0
        E  = 20000.
        EA = 2000000.0
        EI = 210000.0
        w  = -0.1

        params = {'E': E, 'A': EA/E, 'I': EI/E}

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        # create nodes

        nd0 = Node(0.0, 0.0)
        model += nd0

        ndi = nd0
        for i in range(N):
            # nodes
            ndj = Node( 0.0, (i+1)*L/N )
            model += ndj

            # elements
            elem = Frame2D(ndi, ndj, ElasticSection(params))
            elem.setDistLoad(w)
            model += elem

            ndi = ndj

        # define support(s)
        nd0.fixDOF('ux', 'uy')    # horizontal and vertical support bottom end
        ndi.fixDOF('ux')          # horizontal support top end

        # add loads
        # .. load only the upper nodes
        Pcr = np.pi**2 * EI / L**2
        ndi.setLoad((-0.5*Pcr,), ('uy',))

        # show model information
        print(model)

        model.solve(verbose=True)

        model.report()

        model.plot(factor=10.0, filename="frame2_deformed.png")

        model.beamValuePlot("F", filename="frame2_force.png")
        model.beamValuePlot("V", filename="frame2_shear.png")
        model.beamValuePlot("M", filename="frame2_moment.png")

        return



"""
Example
"""
from ...examples.Example import *

from ...domain.System import *
from ...domain.Node import *
from ...elements.Beam2D import *
from ...materials.ElasticSection import *


class ExampleBeam02(Example):

    def docString(self):
        s = """
        Three-span continuous beam under uniform load. 
        
        The system is statically determined and allows for easy validation of 
        calculated deformation, reactions and internal forces.
        
        Author: Peter Mackenzie-Helnwein 
        """
        return s

    def problem(self):
        # initialize a system model
        SpanLengths = [ 8.0 * 12, 10.0 * 12, 8.0 * 12 ]
        Nelems = 2     # number of elements
        params = {'E': 29000., 'A': 5, 'I':50}

        # define load
        w = -1.00

        model = System()

        # meshing parameters
        Xnode  = 0.0
        Ynode  = 0.0
        Offset = 0.0

        # create left node
        nd0 = Node(Xnode, Ynode)
        nd0.fixDOF('ux', 'uy')     # pin support left end
        model += nd0

        # initialization for node and element creation
        ndi = nd0

        for SpanLength in SpanLengths:

            Le = SpanLength / Nelems

            for e in range(Nelems):
                # create next node
                Xnode += Le
                ndj = Node(Xnode, Ynode)
                model += ndj

                # create elements
                elem = Beam2D(ndi, ndj, ElasticSection(params))
                model += elem

                # load the element with a uniform load
                elem.setDistLoad(w)

                # shift one node to the right
                ndi = ndj

            # define support(s)
            ndj.fixDOF('uy')           # roller support right end

            # move on to the next span
            Offset = Xnode

        # done building the model

        # analyze the model
        model.solve()

        # write out report
        model.report()

        # create plots
        model.plot(factor=100., filename="beam02_deformed.png")

        model.beamValuePlot('V', filename="beam02_shear.png")
        model.beamValuePlot('M', filename="beam02_moment.png")




"""
Example
"""
from ...examples.Example import *

from ...domain.System import *
from ...solver.NewtonRaphsonSolver import NewtonRaphsonSolver
from ...domain.Node import *
from ...elements.Truss import *
from ...materials.FiberMaterial import *


class ExampleTruss02(Example):

    def docString(self):
        s = """
        Statically determinate truss bridge.
        
        The system is statically determined and allows for relatively easy validation of 
        calculated deformation, reactions and internal forces.
        
        Author: Peter Mackenzie-Helnwein 
        """
        return s

    def problem(self):
        # initialize a system model
        P = -10.0      # reference load on top nodes
        B = 6.0 * 12   # with of one bay in inches
        H = 8.0 * 12   # height of one bay in inches

        # material model parameters
        params = {'E': 10000., 'A': 3., 'nu': 0.0, 'fy': 1.e30}

        model = System()
        model.setSolver(NewtonRaphsonSolver())

        # create nodes
        nd0 = Node(0.0, 0.0)
        nd1 = Node(  B, 0.0)
        nd2 = Node(2*B, 0.0)
        nd3 = Node(3*B, 0.0)
        nd4 = Node(4*B, 0.0)
        nd5 = Node(0.5*B, H)
        nd6 = Node(1.5*B, H)
        nd7 = Node(2.5*B, H)
        nd8 = Node(3.5*B, H)

        model.addNode(nd0, nd1, nd2, nd3, nd4, nd5, nd6, nd7, nd8)

        # create elements
        model.addElement(Truss(nd0, nd1, FiberMaterial(params)))  # bottom 1
        model.addElement(Truss(nd1, nd2, FiberMaterial(params)))  # bottom 2
        model.addElement(Truss(nd2, nd3, FiberMaterial(params)))  # bottom 3
        model.addElement(Truss(nd3, nd4, FiberMaterial(params)))  # bottom 4

        model.addElement(Truss(nd5, nd6, FiberMaterial(params)))  # upper 1
        model.addElement(Truss(nd6, nd7, FiberMaterial(params)))  # upper 2
        model.addElement(Truss(nd7, nd8, FiberMaterial(params)))  # upper 3

        model.addElement(Truss(nd0, nd5, FiberMaterial(params)))  # up right diag 1
        model.addElement(Truss(nd1, nd6, FiberMaterial(params)))  # up right diag 2
        model.addElement(Truss(nd2, nd7, FiberMaterial(params)))  # up right diag 3
        model.addElement(Truss(nd3, nd8, FiberMaterial(params)))  # up right diag 4

        model.addElement(Truss(nd1, nd5, FiberMaterial(params)))  # up left diag 1
        model.addElement(Truss(nd2, nd6, FiberMaterial(params)))  # up left diag 2
        model.addElement(Truss(nd3, nd7, FiberMaterial(params)))  # up left diag 3
        model.addElement(Truss(nd4, nd8, FiberMaterial(params)))  # up left diag 4

        # define support(s)
        nd0.fixDOF('ux', 'uy')    # horizontal support left end
        nd4.fixDOF('uy')            # vertical support right end

        # add loads
        # .. load only the upper nodes
        nd5.setLoad((P,), ('uy',))
        nd6.setLoad((P,), ('uy',))
        nd7.setLoad((P,), ('uy',))
        nd8.setLoad((P,), ('uy',))

        model.plot(factor=1.,  filename="truss02_undeformed.png", title="Undeformed System")

        # analyze the model
        model.solve()

        # write out report
        model.report()

        # create plots
        model.plot(factor=20.,  filename="truss02_deformed.png")
        model.beamValuePlot('f',filename="truss02_forces.png")




import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splev, splrep

from .Mesher import *
from ..domain import Node, System

class CurveMesher(Mesher):
    """
    Mesher for beams and frame components.
    """

    def __init__(self, model, *pts):
        super(CurveMesher, self).__init__(model, *pts)

        if len(pts) < 2:
            msg = "{} requires at least 2 points but only {} given".format(self.__class__.__name__, len(pts))
            raise TypeError(msg)

        self.npts    = len(pts)
        self.ndim    = len(pts[0])
        self.support = np.array(pts)

    def mesh(self, Ne, element_type, material, **kwargs):
        """
        mesh a curve in 2d/3d using any kind of line element.

        The given set of points are interpolated using B-splines

        :param int Ne: number of elements along the curve
        :param element_type: compatible element type (BEam, Frame2D, ...)
        :param material: a 1D material object or section model
        :returns (nodes, elements):
        :var nodes: list of created :py:class:`Node` objects
        :var elements: list of created (subclass of) :py:class:`Element` objects
        """
        elements = []
        nodes    = []

        ndeg = np.min( [self.npts-1, 3] )
        ndim = self.ndim

        # source points local coordinates
        s = np.linspace(0, self.npts-1, self.npts)
        # target points local coordinates
        t = np.linspace(0, self.npts-1, Ne+1)

        interpolated = []

        for j in range(ndim):
            xsource = self.support[:,j]
            spl = splrep(s, xsource, k = ndeg)
            xtarget = splev(t, spl)
            interpolated.append(xtarget)

        fem_points = np.array(interpolated).T

        lastnode = None
        for coords in fem_points:
            node = Node(*coords)
            nodes.append(node)
            if lastnode:
                elements.append(element_type(lastnode, node, deepcopy(material)))
            lastnode = node

        if self.model and isinstance(self.model, System):
            self.model.addNode(*nodes)
            self.model.addElement(*elements)

        return (nodes, elements)

    def _create_doc(self):
        """
        create text and figure for the sphinx documentation
        """
        doc_string = """
        .. code:: python
        
            model = System()
            mesher = CurveMesher(model, (0,0),(1.5,.25),(2,1),(3.,1.5))
            mesher.mesh(10, Frame2D, ElasticSection())
        
        .. figure:: CurveMesher01.png
            :alt: CurveMesher
            :align: center
            
            Meshing a curve with 2-noded elements.
            
        """
        print(doc_string)

        points = [(0,0),(1.5,.25),(2,1),(3.,1.5)]
        support = np.array( points )
        npts = len(points)
        Ne = 10

        deg = np.min( [npts-1, 3] )

        s = np.linspace(0, npts-1, npts)
        x = support[:,0]
        y = support[:,1]
        splx = splrep(s, x, k = deg)
        sply = splrep(s, y, k = deg)

        s2 = np.linspace(0, self.npts-1, Ne+1)
        x2 = splev(s2, splx)
        y2 = splev(s2, sply)
        plt.plot(x, y, 'bo', ms=10, label="Input points")
        plt.plot(x2, y2, '-ro', label="Created mesh")
        plt.gca().set_aspect('equal')
        plt.gca().set_aspect('equal')
        plt.legend()
        plt.savefig("CurveMesher01.png", bbox_inches='tight')
        #plt.show()






from .Material import *

class PlaneStress(Material):
    """
    class: representing a 2d Plane Stress Material

    """

    def __init__(self, params={'E':1.0, 't':1.0, 'nu':0.0, 'fy':1.0e30}):
        super().__init__(params = params)

        self._type = self.PLANESTRESS | self.PLASTIC

        if 't' not in self.parameters:
            self.parameters['t'] = 1.0

        # initialize strain
        self.total_strain   = np.zeros(3)
        self.plastic_strain = np.zeros(3)
        self.setStrain({'xx':0.0, 'yy':0.0, 'zz':0.0, 'xy':0.0, 'yz':0.0, 'zx':0.0})

    def getThickness(self):
        if 't' in  self.parameters:
            return self.parameters['t']
        else:
            return 1.0

    def updateState(self):
        """
        update state for a user provided axial strain value

        :param eps:  strain or strain tensor
        :return: n/a
        """
        eps = np.array([self.strain['xx'], self.strain['yy'], self.strain['xy']])

        # update stress state
        E  = self.parameters['E']
        t  = self.parameters['t']
        nu = self.parameters['nu']
        fy = self.parameters['fy']

        # default consistency parameter
        gamma = 0.0

        self.Et = E*t/(1. - nu*nu) * np.array([[1.,nu,0.],[nu,1.,0.],[0.,0.,(1.-nu)/2.]])

        Cinv = 1/(E*t) * np.array([[1.,-nu,0.],[-nu,1.,0.],[0.,0.,2.*(1.+nu)]])
        Phi  = np.array([[2.,-1.,0.],[-1.,2.,0.],[0.,0.,6.]])

        # elastic predictor
        stress = self.Et @ ( eps - self.plastic_strain )

        self.strain['zz'] = -self.parameters['nu'] * (eps[0] + eps[1])  # elastic thickness strain

        # check yield condition
        (sxx, syy, sxy) = stress
        f = stress @ Phi @ stress / 2. - (t*fy)**2

        gamma = 0.0
        Xi = self.Et

        # plastic corrector as needed
        if f >= 0.0:
            # loop
            r = Phi @ stress
            Xixr = Xi @ r
            gamma += f / (r @ Xixr)
            Xi = np.linalg.inv(Cinv + gamma * Phi)
            stress = Xi @ ( eps - self.plastic_strain )
            self.Et  = Xi

            self.strain['zz']  = -0.5 * (self.plastic_strain[0] + self.plastic_strain[1])
            self.strain['zz'] += -nu * (stress[0] + stress[1]) / E

            print("material entering plastic state")

        #print(4*'{:12.8e}  '.format(eps, f, self.plastic_strain, self.sig ))
        self.sig    = stress
        self.stress = {'xx':stress[0], 'yy':stress[1], 'zz':0.0, 'xy':stress[2], 'yz':0.0, 'zx':0.0}

    def getStrain(self):
        return self.strain

    def converged(self):
        # update state now that the global analysis has converged
        E  = self.parameters['E']
        t  = self.parameters['t']
        nu = self.parameters['nu']

        Cinv = 1 / (E * t) * np.array([[1., -nu, 0.], [-nu, 1., 0.], [0., 0., 2. * (1. + nu)]])
        self.plastic_strain = self.total_strain - Cinv @ self.sig



import numpy as np

from .Transformation import *

class SolidTransformation(Transformation):
    r"""
    A Transformation including only displacement dofs (no rotational dofs)
    """

    def __init__(self, dir1=None, dir2=None, axis=None):
        super().__init__(dir1, dir2, axis)



"""
Example
"""
from ...examples.Example import *

from ...domain.System import *
from ...domain.Node import *
from ...elements.LinearTriangle import *
from ...materials.PlaneStress import *


class ExamplePlate01(Example):

    def docString(self):
        s = """
    ## A square patch made of two triangular plate elements

    Basic implementation test with all prescribed displacements.  
    Testing the internal force computation. 

    v=-5   v=-5
     |     |
     v     v
     3-----2 -> u=5
     |\  b | >
     | \   | >
     |  \  | > (w = 1.0)
     |   \ | >
     | a  \| >
     0-----1 -> u=5
    
    width:  10.
    height: 10.
    
    Material parameters: St. Venant-Kirchhoff, plane stress
        E  = 10.0
        nu =  0.30
        t  =  1.0
        
    Element loads:
        node 0: [ 0.0, 0.0]
        node 1: [ 5.0, 0.0]
        node 2: [ 5.0, 0.0]
        node 3: [ 0.0, 0.0]
    
    Green Lagrange strain:
        eps_XX = 0.5 * ((1.5)^2 - 1)      =  0.625
        eps_YY = 0.5 * ((0.5)^2 - 1)      = -0.375
        eps_XY = eps_YX                   =  0.000
        eps_ZZ = - nu * (eps_XX + eps_YY) = -0.075
        
    2nd Piola-Kirchhoff stress:
        D = E t/(1 - nu^2) = 10.989
        S_XX = (10.989) * ((0.625) + (0.30)(-0.375)) =  5.632
        S_YY = (10.989) * ((-0.375) + (0.30)(0.625)) = -2.060
        S_XY = S_YX = S_ZZ                           =  0.000
        
    Author: Peter Mackenzie-Helnwein 
    """
        return s

    def problem(self):

        params = dict(
            E  = 10.,   # Young's modulus
            nu = 0.3,   # Poisson's ratio
            t  = 1.0,   # thickness of the plate
            fy = 1.e30  # yield stress
        )

        a = 10.     # length of the plate in the x-direction
        b = 10.     # length of the plate in the y-direction

        model = System()

        nd0 = Node( 0.0, 0.0)
        nd1 = Node(   a, 0.0)
        nd2 = Node(   a,   b)
        nd3 = Node( 0.0,   b)

        model.addNode(nd0, nd1, nd2, nd3)

        elemA = LinearTriangle(nd0, nd1, nd3, PlaneStress(params))
        elemB = LinearTriangle(nd2, nd3, nd1, PlaneStress(params))

        model.addElement(elemA, elemB)

        elemB.setSurfaceLoad(face=2, w=1.0)

        model.plot(title="Undeformed system", filename="plate01_undeformed.png")

        model.setLoadFactor(1.0)

        nd0.setDisp( [0.0, 0.0] )
        nd1.setDisp( [5.0, 0.0] )
        nd2.setDisp( [5.0,-5.0] )
        nd3.setDisp( [0.0,-5.0] )

        elemA.updateState()
        elemB.updateState()

        model.report()

        model.plot(factor=1.0, filename="plate01_deformed.png")




from .Recorder import *
from .Record import *

class MaterialRecorder(Recorder):
    """
    Recorder class for state variables at integration points

    .. warning::

        Use this sparely and only for very few elements since is can generate
        **lots of mostly unnecessary data**.

    """

    def __init__(self, **kwargs):
        super(MaterialRecorder, self).__init__(**kwargs)

        if 'label' in kwargs:
            template = kwargs['label'] + ":{}"
        else:
            template = "material:{}"

        for var in self.materialvars:
            lbl = template.format(var)
            self.data[var] = Record(key=var, label=lbl)




__all__ = (
    'Node',
    'System',
    'Transformation',
    'CosseratTransformation',
    'BeamTransformation',
    'Beam2dTransformation',
    'FrameTransformation',
    'Frame2dTransformation',
    'SolidTransformation',
    'Solid2dTransformation'
)

from .System                import System
from .Node                  import Node
from .Transformation        import Transformation
from .FrameTransformation   import FrameTransformation
from .Frame2dTransformation import Frame2dTransformation
from .BeamTransformation    import BeamTransformation
from .Beam2dTransformation  import Beam2dTransformation
from .SolidTransformation   import SolidTransformation
from .Solid2dTransformation import Solid2dTransformation



import numpy as np
import scipy.sparse as scs
import scipy.sparse.linalg as spla

from ..solver.NewtonRaphsonSolver import NewtonRaphsonSolver

class SparseSolver(NewtonRaphsonSolver):
    """
    A sparse solver version for load-controlled nonlinear analysis.
    """

    def __init__(self):
        super(SparseSolver, self).__init__()

    def solve(self, max_steps=10, verbose=False, **kwargs):
        """
        :param max_step: maximum number of iterations (int)
        :param verbose: set to :code:`True` for additional information
        """
        TOL = self.TOL

        if 'tol' in kwargs:
            TOL = kwargs['tol']
        if 'tolerance' in kwargs:
            TOL = kwargs['tolerance']

        for k in range(max_steps):

        # compute force vector and tangent stiffness
            self.assemble()

            # we are using , force_only=False and simply reuse self.Kt from that run (!)
            normR = self.checkResiduum(verbose, force_only=False)

            if normR < TOL:
                # we achieved convergence
                #
                # now broadcast that we can switch to this converged state
                self.on_converged()

                # time to add the information to the recorded data
                if self.record:
                    self.recordThisStep()

                break

            # Solve for equilibrium
            self.solveSingleStep()

        print('+')

        if normR > TOL:
            # we failed to converge
            #
            # back to safety: revert to the last converged step
            self.revert()

        return normR

    def solveSingleStep(self):
        """
        Helper function performing a single solution of the linearized system

        Called by **solve()**. (internal use only)
        """

        # are we doing displacement control?
        if self.hasConstraint:

            # solve for displacement update: a single Newton step
            ##dQ = np.linalg.solve(self.Kt, np.stack([self.R, self.P]).T)# LU factorization of Kt
            solve = spla.factorized(self.Kt)

            # solve for displacement update: a single Newton step
            dq0 = solve(self.R)
            dq1 = solve(self.P)

            if self.useArcLength:
                # arc-length control
                delU  = self.sysU - self.lastConverged['U']
                dload = self.loadfactor - self.lastConverged['lambda']
                # g = self.arclength2 - delU@delU - self.alpha * dload*dload * self.P@self.P

                denum = 2.*np.dot(dq1, delU) + 2.*self.alpha * dload * self.P@self.P

                dlam = ( self.g - 2.*np.dot(dq0, delU) ) / denum

            else:
                # displacement control
                # g = (self.targetU - np.dot(self.sysU, self.en))

                dlam = self.g - np.dot(dq0, self.en)
                dlam /= np.dot(dq1, self.en)

            self.loadfactor += dlam
            dU = dq0 + dlam * dq1

        else:
            dU = spla.spsolve(self.Kt, self.R)

        # update nodal displacements
        for node in self.nodes:
            idxK = node.start + np.arange(node.ndofs)
            node._updateDisp(dU[idxK])

    def assemble(self, force_only=False):
        """
        A general assembler for mixed element types.

        This method will build the out-of-balance force vector (residuum :math:`{\\bf R}`)
        and the tangent stiffness matrix (:math:`{\\bf K}_t`) used by most solvers.

        Specialized solvers may overload this method.

        .. note::

            The solver will apply the global load factor to the reference load returned by nodes
            and elements.  While nodes and elements are aware of that load factor, they shall
            apply it **only if asked explicity** by passing :code:`apply_load_factor=True` to the
            respective access functions.

        :param force_only: set to **True** if only the residual force needs to be assembled
        """

        # compute size parameters
        ndof = 0
        for node in self.nodes:
            node.setStart(ndof)
            ndof += node.ndofs

        for constraint in self.constraints:
            constraint.setStart(ndof)
            ndof += constraint.countConditions()

        self.sdof = ndof  # number of system d.o.f.s

        Rsys = np.zeros(ndof)

        rows = []
        cols = []
        data = []

        # assemble loads
        for node in self.nodes:
            if node.hasLoad():
                idx = node.start + np.arange(node.ndofs)
                Rsys[idx] += node.getLoad() * self.loadfactor

        # Element Loop: assemble element forces and stiffness
        for element in self.elements:
            Fe = element.getForce()     # Element State Update occurs here
            Pe = element.getLoad()      # Element State Update occurs here
            for (i,ndI) in enumerate(element.nodes):
                idxK = ndI.start + np.arange(ndI.ndofs)
                if isinstance(Pe[i], np.ndarray):
                    Rsys[idxK] -= Fe[i] - self.loadfactor * Pe[i]
                else:
                    Rsys[idxK] -= Fe[i]
                if not force_only:
                    for (j,ndJ) in enumerate(element.nodes):
                        idxM = ndJ.start + np.arange(ndJ.ndofs)
                        KIJ = element.Kt[i][j]
                        II,JJ = KIJ.shape
                        for p in range(II):
                            for q in range(JJ):
                                rows.append(idxK[p])
                                cols.append(idxM[q])
                                data.append(KIJ[p][q])

        # apply boundary conditions
        if not force_only:
            for node in self.nodes:
                for dof in node.dofs:
                    if node.isFixed(dof):
                        idx = node.start + node.dofs[dof]
                        Rsys[idx]      = 0.0
                        rows.append(idx)
                        cols.append(idx)
                        data.append(1.0e30)

        self.R  = Rsys

        KtS = scs.coo_array((data, (rows, cols)), shape=(self.sdof,self.sdof))
        self.Kt = scs.csc_array(KtS)



"""
Example
"""
from ...examples.Example import *

from ...domain.System import *
from ...domain.Node import *
from ...elements.Beam2D import *
from ...materials.ElasticSection import *


class ExampleBeam01(Example):

    def docString(self):
        s = """
        Single span beam under uniform load. 
        
        The system is statically determined and allows for easy validation of 
        calculated deformation, reactions and internal forces.
        
        Author: Peter Mackenzie-Helnwein 
        """
        return s

    def problem(self):
        # initialize a system model
        SpanLength = 10.0 * 12
        w =  -1.0   # distributed load (positive if acting in local y-direction
        P =   -40.0   # center point load (uses global system)

        Nelems = 4    # number of elements
        params = {'E': 29000., 'A': 4.7, 'I':103}

        model = System()

        # meshing parameters
        Le = SpanLength / Nelems
        Xnode = 0.0
        Ynode = 0.0

        # create left node
        nd0 = Node(Xnode, Ynode)
        model += nd0

        ndP = None

        # initialization for node and element creation
        ndi = nd0

        for e in range(Nelems):
            # create next node
            Xnode += Le
            ndj = Node(Xnode, Ynode)
            model += ndj

            # remember center node for loading
            if Xnode <= SpanLength/2:
                ndP = ndj

            # create elements
            elem = Beam2D(ndi, ndj, ElasticSection(params))
            model += elem

            # load the element
            elem.setDistLoad(w)

            # shift one node to the right
            ndi = ndj

        # define support(s)
        nd0.fixDOF('ux', 'uy')     # pin support left end
        ndj.fixDOF('uy')           # roller support right end

        # add point loads
        # .. load only the center node
        if ndP:
            ndP.setLoad([0.0, P], ('ux', 'uy'))

        # analyze the model
        model.solve()

        # write out report
        model.report()

        # create plots
        model.plot(factor=10., filename="beam01_deformed.png")

        model.beamValuePlot('V', filename="beam01_shear.png")
        model.beamValuePlot('M', filename="beam01_moment.png")




from .AbstractPlotter import *


class Plotter(AbstractPlotter):
    """
    class: representing a Plotter object
    """

    def __init__(self):
        super(Plotter, self).__init__()
        self.vertices  = []
        self.lines     = []

    def setMesh(self, vert, lines):
        """

        :param vert:
        :param lines:
        """
        self.vertices = np.array(vert)
        self.lines    = np.array(lines)

    def setDisplacements(self, disp):
        """

        :param disp:
        """
        self.disp = np.array(disp)

    def setValues(self, vals):
        """

        :param vals:
        """
        self.values = np.array(vals)

    def setReactions(self, R):
        """

        :param R: list or tuple of nodal force vectors
        """
        self.reactions = R

    def displacementPlot(self, file=None, **kwargs):
        """
        Create a deformed system plot

        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param file: filename (str)
        """
        if len(self.vertices[0]) == 3:
            fig = plt.figure(figsize=(10, 10))
            axs = fig.gca(projection='3d')

            # plot the undeformed lines
            for line in self.lines:
                vert0 = self.vertices[line[0]]
                vert1 = self.vertices[line[1]]
                x = [vert0[0], vert1[0]]
                y = [vert0[1], vert1[1]]
                z = [vert0[2], vert1[2]]
                axs.plot(x, y, z, '-k', lw=2)

            # plot the deformed lines
            if len(self.disp) == len(self.vertices):
                for line in self.lines:
                    vert0 = self.vertices[line[0]].copy()
                    vert1 = self.vertices[line[1]].copy()
                    vert0 += self.disp[line[0]]
                    vert1 += self.disp[line[1]]
                    x = [vert0[0], vert1[0]]
                    y = [vert0[1], vert1[1]]
                    z = [vert0[2], vert1[2]]
                    axs.plot(x, y, z, '-r', lw=3)

            if self.reactions != []:
                self.addForces(axs)

            self.set_axes_equal(axs)
            #axs.set_aspect('equal')
            #axs.set_axis_off()

        else:
            fig, axs = plt.subplots()

            # plot the undeformed lines
            for line in self.lines:
                vert0 = self.vertices[line[0]]
                vert1 = self.vertices[line[1]]
                x = [vert0[0], vert1[0]]
                y = [vert0[1], vert1[1]]
                axs.plot(x, y, '-k', lw=2)

            # plot the deformed lines
            if len(self.disp) == len(self.vertices):
                for line in self.lines:
                    vert0 = self.vertices[line[0]].copy()
                    vert1 = self.vertices[line[1]].copy()
                    vert0 += self.disp[line[0]]
                    vert1 += self.disp[line[1]]
                    x = [vert0[0], vert1[0]]
                    y = [vert0[1], vert1[1]]
                    axs.plot(x, y, '-r', lw=3)

            if self.reactions != []:
                self.addForces(axs)

            axs.set_aspect('equal')
            axs.set_axis_off()

        plt.show()

    def valuePlot(self, deformed=False, file=None, **kwargs):
        r"""
        Create a plot using colors to identify magnitude of internal force.

        .. list-table:: known variable_name
            :header-rows: 1

            * - keyword
              - description
            * - 'ux', 'uy', 'uz'
              - displacement component in the `x`, `y` or `z` direction
            * - 'rx', 'ry', 'rz'
              - rotation around the `x`, `y` or `z` axis
            * - 'sxx', 'syy', 'szz'
              - normal stress :math:`\sigma_{xx}`, :math:`\sigma_{yy}`, :math:`\sigma_{zz}`
            * - 'sxy', 'syz', 'szx'
              - shear stress :math:`\sigma_{xy}`, :math:`\sigma_{yz}`, :math:`\sigma_{zx}`
            * - 'epsxx', 'epsyy', 'epszz'
              - normal strain :math:`\varepsilon_{xx}`, :math:`\varepsilon_{yy}`, :math:`\varepsilon_{zz}`
            * - 'epsxy', 'epsyz', 'epszx'
              - engineering shear strain :math:`\gamma_{xy}=2\varepsilon_{xy}`,
                :math:`\gamma_{yz}=2\varepsilon_{yz}`,
                :math:`\gamma_{zx}=2\varepsilon_{zx}`
            * - 'T'
              - temperature
            * - 'qx', 'qy', 'qz'
              - `x`, `y` or `z` component of the temperature gradient, :math:`q_i = \partial T/\partial x_i`


        If **file** is given, store the plot to that file.
        Use proper file extensions to indicate the desired format (.png, .pdf)

        :param deformed: True | **False**
        :param file: filename (str)
        """
        fig, axs = plt.subplots()

        # plot the lines
        segments = []

        if len(self.disp) == len(self.vertices):
            for line in self.lines:
                vert0 = self.vertices[line[0]].copy()  # we need a copy since we will be modifying this in the lines below
                vert1 = self.vertices[line[1]].copy()  # we need a copy since we will be modifying this in the lines below
                if deformed:
                    vert0 += self.disp[line[0]]   # it's this += that modifies the vertices if we don't use a copy
                    vert1 += self.disp[line[1]]   # it's this += that modifies the vertices if we don't use a copy
                #x = [vert0[0], vert1[0]]
                #y = [vert0[1], vert1[1]]
                #axis.plot(x,y,'-r',lw=3)
                segments.append(np.array([vert0, vert1]))

        # Create a continuous norm to map from data points to colors
        lc = LineCollection(np.array(segments), cmap='rainbow')
        # Set the values used for colormapping
        lc.set_array(self.values)
        lc.set_linewidth(3)
        line = axs.add_collection(lc)
        fig.colorbar(line, ax=axs)

        if self.reactions != []:
            self.addForces(axs)

        axs.set_aspect('equal')
        axs.set_axis_off()

        plt.autoscale(enable=True, axis='x', tight=False)
        plt.autoscale(enable=True, axis='y', tight=False)
        plt.show()

    def addForces(self, axs):
        """
        add nodal forces to the plot shown in **axs**

        :param axs: axis on which to plot
        """
        if len(self.reactions) == len(self.vertices):
            Fx = []
            Fy = []
            X = []
            Y = []
            for (point, force) in zip(self.vertices, self.reactions):
                if np.linalg.norm(force) > 1.0e-3:
                    X.append(point[0])
                    Y.append(point[1])
                    Fx.append(-force[0])
                    Fy.append(-force[1])

            axs.quiver(X,Y, Fx, Fy, color='green')


if __name__ == "__main__":
    # testing the plotter
    plotter = Plotter()



"""
==============================================
4-node quadrilateral - small displacement
==============================================
complete bi-linear interpolation
with selective reduced integration for shear

.. code::

         1
      x     y
        x*y


"""
import numpy as np
from copy import deepcopy

from ..Element import *
from ...domain.Node import *
from ...utilities import QuadIntegration, QuadShapes

class ReducedIntegrationQuad(Element):
    """
    class: representing a plane 4-node quadrilateral

    This element works as 2D plate, loaded in-plane, and as a 3D membrane element.

    * For 2D plate behavior, define nodes as two-dimensional nodes
    * For 3D membrane behavior, define nodes as three-dimensional nodes
    """

    def __init__(self, node0, node1, node2, node3, material, label=None):
        super(ReducedIntegrationQuad, self).__init__((node0, node1, node2, node3), material, label=label)
        self.element_type = DrawElement.QUAD
        self.createFaces()

        if node0.getPos().size == 3:
            dof_list = ('ux','uy','uz')
            ndof = 3
        elif node0.getPos().size == 2:
            dof_list = ('ux','uy')
            ndof = 2
        else:
            raise TypeError("dimension of nodes must be 2 or 3")

        self._requestDofs(dof_list)

        self.distributed_load = [0.0, 0.0, 0.0, 0.0]   # face loads along the perimeter of the element
        self.force    = 0.0
        self.Forces   = [ np.zeros(ndof) for k in range(len(self.nodes)) ]
        self.Kt       = [ [ np.zeros(ndof) for k in range(len(self.nodes)) ] for m in range(len(self.nodes)) ]
        self.ndof = ndof

        self.material  = []   # one material object per integration point
        self.stress    = []   # hold gauss-point stress (1st Piola-Kirchhoff stress)
        self.J         = []   # jacobian at integration point

        self.Grad      = []   # derivative of shape functions with respect to global coords

        X  = np.array([ node.getPos() for node in self.nodes ])

        ## initialization step

        interpolation = QuadShapes()

        # reduced integration
        # -------------------------
        xi = (0.0, 0.0)   # gauss point at the center
        wi = 4.0          # weight at the center

        dphi_ds = interpolation.shape(1, *xi, n=(1, 0))
        dphi_dt = interpolation.shape(1, *xi, n=(0, 1))
        Grad = np.vstack((dphi_ds, dphi_dt))

        # reference configuration

        DPhi0 = (Grad @ X).T
        self.J0 = np.linalg.det(DPhi0)   # material model already includes thickness

        # dual base (contra-variant)
        self.Grad0 = np.linalg.inv(DPhi0).T @ Grad

        self.material0 = deepcopy(material)
        self.stress0   = {}

        # full integration
        # -------------------------
        integrator = QuadIntegration(order=2)
        xis, wis = integrator.parameters()

        gpt = 0

        for xi, wi in zip(xis, wis):

            dphi_ds = interpolation.shape(1, *xi, n=(1,0))
            dphi_dt = interpolation.shape(1, *xi, n=(0,1))
            Grad = np.vstack((dphi_ds, dphi_dt))

            # reference configuration
            # -------------------------

            #
            # $ D\Phi_0 $
            #
            DPhi0 = (Grad @ X).T
            self.J.append( np.linalg.det(DPhi0) )  # material model already includes thickness

            # dual base (contra-variant)
            self.Grad.append( np.linalg.inv(DPhi0).T @ Grad)

            self.material.append(deepcopy(material))
            self.stress.append({})
            gpt += 1



    def __str__(self):
        s = super(ReducedIntegrationQuad, self).__str__()
        for igpt, material in enumerate(self.material):
            s += "\n    strain ({}): xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(igpt,**material.getStrain())
            s += "\n    stress ({}): xx={xx:.3e} yy={yy:.3e} xy={xy:.3e} zz={zz:.3e}".format(igpt,**material.getStress())
        if np.array(self.distributed_load).any():
            s += "\n    element forces added to node:"
            for i, P in enumerate(self.Loads):
                Pi = np.array(P) * self.loadfactor
                s += "\n        {}: {}".format(self.nodes[i].getID(), Pi)
        return s

    def setSurfaceLoad(self, face, pn, ps=0):
        """
        .. list-table::
            :header-rows: 1

            * - face ID
              - nodes defining that face
            * - 0
              - :py:obj:`node 0` to :py:obj:`node 1`
            * - 1
              - :py:obj:`node 1` to :py:obj:`node 2`
            * - 2
              - :py:obj:`node 2` to :py:obj:`node 3`
            * - 3
              - :py:obj:`node 3` to :py:obj:`node 0`


        :param face: face ID for the laoded face
        :param pn: magnitude of distributed normal load per area. Tension on a surface is positive.
        :param ps: magnitude of distributed shear load per area. Positive direction is defined as shown in the above table.
        """
        if face >= 0 and face <= 3:
            self.faces[face].setLoad(pn, ps)

    def resetLoads(self):
        super(ReducedIntegrationQuad, self).resetLoads()

    def updateState(self):

        # initialization step
        integrator = QuadIntegration(order=2)

        nnds = len(self.nodes)
        ndof = self.ndof       # mechanical element

        self.Forces = [ np.zeros(ndof) for k in range(nnds) ]
        R  = [ np.zeros((ndof,ndof)) for i in range(nnds) ]
        Kt = [ [ np.zeros((ndof,ndof)) for i in range(nnds) ] for j in range(nnds) ]

        # create array of deformed nodal coordinates
        xt = np.array([ node.getDeformedPos() for node in self.nodes ])

        # interpolation = QuadShapes()   # we are doing that and the isoparametric transformation in the constructor

        # reduced integration
        # -------------------------
        xi = (0.,0.)
        wi = 4.0

        Grad = self.Grad0  # pre-computed in __init__
        wi *= self.J0  # J includes the thickness of the plate

        # spatial configuration
        # -------------------------

        # deformation gradient
        F = (Grad @ xt).T

        # compute small strain tensor
        eps = 0.5 * (F + F.T) - np.eye(self.ndof)

        # update the material state
        strain0 = {'xx': eps[0, 0], 'yy': eps[1, 1], 'xy': eps[0, 1] + eps[1, 0]}

        self.material0.setStrain(strain0)

        # 2nd Piola-Kirchhoff stress
        stress = self.material0.getStress()

        S = np.array([[stress['xx'], stress['xy']], [stress['xy'], stress['yy']]])

        # store stress for reporting
        self.stress0 = {'xx': S[0, 0], 'xy': S[0, 1], 'yx': S[1, 0], 'yy': S[1, 1]}

        Ones = np.eye(self.ndof)
        # compute kinematic matrices
        BI = [np.array([Grad[0, K] * Ones[:, 1] + Grad[1, K] * Ones[:, 0]])   # the XY component is XY + YX ("gammaXY = 2 epsXY")
              for K in range(nnds)]

        # internal forces
        ##for i, force in enumerate(self.Forces):
        ##    force += S @ Grad[:, i] * wi  # wi already includes J

        # tangent stiffness
        Ct = self.material0.getStiffness() * wi

        for I, Bi in enumerate(BI):
            GCti = Bi.T * Ct[2,2]
            for J, Bj in enumerate(BI):
                Kt[I][J] += GCti @ Bj


        # full integration
        # -------------------------
        xis, wis = integrator.parameters()

        gpt = 0

        for xi, wi in zip(xis, wis):

            # reference configuration
            # -------------------------

            Grad = self.Grad[gpt]  # pre-computed in __init__
            wi *= self.J[gpt]      # J includes the thickness of the plate

            # spatial configuration
            # -------------------------

            # deformation gradient
            F = (Grad @ xt).T

            # compute small strain tensor
            eps = 0.5 * ( F + F.T ) - np.eye(self.ndof)

            # update the material state
            # ... replacing the shear strain at the gauss-points by the center shear strain
            #strain = {'xx':eps[0,0], 'yy':eps[1,1], 'xy':eps[0,1]+eps[1,0]}
            strain = {'xx':eps[0,0], 'yy':eps[1,1], 'xy':strain0['xy']}

            self.material[gpt].setStrain(strain)

            # 2nd Piola-Kirchhoff stress
            stress = self.material[gpt].getStress()

            S = np.array( [[stress['xx'],stress['xy']],[stress['xy'],stress['yy']]] )

            # store stress for reporting
            self.stress[gpt] = {'xx':S[0,0], 'xy':S[0,1], 'yx':S[1,0], 'yy':S[1,1]}

            Ones = np.eye(self.ndof)
            # compute kinematic matrices
            BI = [ np.array([ Grad[0,K]*Ones[:,0],                         # the XX component
                              Grad[1,K]*Ones[:,1]])                        # the YY component
                   for K in range(nnds) ]

            # internal forces
            for i, force in enumerate(self.Forces):
                force += S @ Grad[:,i] * wi  # wi already includes J

            # tangent stiffness
            Ct = self.material[gpt].getStiffness() * wi

            for I, Bi in enumerate(BI):
                GCti = Bi.T @ Ct[0:2,0:2]
                for J, Bj in enumerate(BI):
                    Kt[I][J] += GCti @ Bj

            # on to the next integration point
            gpt += 1

        self.Kt = Kt

    def computeSurfaceLoads(self):
        """
        compute surface loads using faces

        This method should be called during :py:meth:`updateState()` by every
        element supporting surface loads

        """
        self.Loads = [ np.zeros_like(self.Forces[I]) for I in range(len(self.nodes)) ]

        for I, face in enumerate(self.faces):
            loads = face.computeNodalForces()

            # indexing
            J = I+1
            if J>3:
                J -= 4

            # add to element load vectors
            self.Loads[I] += loads[0]
            self.Loads[J] += loads[1]
            if loads.shape[0]>2:
                numNodes = len(self.nodes)
                if numNodes == 8 or numNodes == 9:
                    K = I + 4
                else:
                    msg = "Force data provided from Face2D inconsistent with element data"
                    raise TypeError(msg)

                self.Loads[K] += loads[2]

    def getStress(self):
        return self.Stress



